from dataclasses import replace
from pathlib import Path
from typing import TYPE_CHECKING, override

from ai_parade._toolkit.converters._pytorch_base import PyTorchConvertBase
from ai_parade._toolkit.data.ImageInput import ImageInput
from ai_parade._toolkit.enums import (
	ExactModelFormat,
	HardwareAcceleration,
	ModelFormat,
	Quantization,
)

from .converter import ConversionResult, Converter

if TYPE_CHECKING:
	import torch  # type: ignore


class Executorch(Converter):
	"""
	https://docs.pytorch.org/executorch/main/using-executorch-export.html
	"""

	def __init__(
		self,
		quantization: Quantization,
		hw_acceleration: HardwareAcceleration,
		soc: tuple[str, ...] | None,
	):
		self._quantization = quantization
		self._hw_acceleration = hw_acceleration
		self._soc = soc

	chip_compatibility_lists: list[tuple[str, tuple[str, ...]]] = [
		("SM8450", ("SM8450",)),
		("SM8550", ("SM8550", "SM8550-AC")),
		("SM8650", ("SM8650", "SM8650-AC")),
		("SM8750", ("SM8750", "SM8750-AC")),
	]

	@property
	@override
	def conversion(self):
		return (
			ExactModelFormat(ModelFormat.PyTorch),
			ExactModelFormat(
				ModelFormat.ExecutorTorch,
				self._quantization,
				self._hw_acceleration,
				self._soc,
			),
		)

	@override
	def create(self, options: Converter.Options):
		super().create(options)
		return self.ExecutorchConverter(self.conversion, options)

	class ExecutorchConverter(PyTorchConvertBase):
		@override
		def _export_call(self, params: PyTorchConvertBase.Params):
			(
				model,
				sample_inputs,
				output_weights_path,
				default_results,
				model_metadata,
				*_,
			) = params

			import torch  # type: ignore
			from executorch.exir import to_edge_transform_and_lower  # type: ignore
			from torchao.quantization.pt2e.quantize_pt2e import (  # type: ignore
				convert_pt2e,
				prepare_pt2e,
			)

			image_input = default_results.image_input
			if self.conversion[1].quantization == Quantization.float16:
				image_input = replace(image_input, dataType=ImageInput.DataType.float16)

			# todo
			# model = model.to(memory_format=torch.channels_last)  # pyright: ignore[reportCallIssue]
			# sample_inputs = (sample_inputs[0].to(memory_format=torch.channels_last),)

			# ?TODO: add support for dynamic shapes,
			# problem is the executorch doesn't support unbound dims (i.e. [1, inf])
			# and for some reason it complains with sample output batch dim set to 1:
			# "You marked batch_dim as dynamic but your code specialized it to be a constant (1)"
			dynamic_shapes = ()  # ({0: torch.export.Dim("batch_dim", min=1, max=8)},)
			match self.conversion[1].hw_acceleration_restriction:
				case HardwareAcceleration.XNNPACK:
					partitioner, quantizer = self.xnnpack_setup()
				case HardwareAcceleration.Vulkan:
					partitioner, quantizer = self.vulkan_setup()
				case HardwareAcceleration.QNN:
					partitioner, quantizer = self.qnn_setup()
				case _:
					raise NotImplementedError(
						f"{self.conversion[1].hw_acceleration_restriction} is not supported"
					)
			match self.conversion[1].quantization:
				case Quantization.none:
					pass
				case Quantization.float16:
					model = model.half()
				case Quantization.int8:
					assert self.options.calibration_data is not None

					training_ep = torch.export.export(
						model, sample_inputs, dynamic_shapes=dynamic_shapes
					).module()
					prepared_model = prepare_pt2e(training_ep, quantizer)

					# calibration
					for cal_sample, _, _ in self.options.calibration_data:
						prepared_model(torch.from_numpy(cal_sample))

					model = convert_pt2e(prepared_model)

			if (
				self.conversion[1].hw_acceleration_restriction
				== HardwareAcceleration.QNN
			):
				models = self.convert_qnn(model, sample_inputs, output_weights_path)
				return [
					ConversionResult(
						format=replace(self.conversion[1], chip_compatibility=[soc]),
						image_input=image_input,
						path=model,
						output=model_metadata.output,
					)
					for model, soc in models
				]
			else:
				assert partitioner is not None
				executorch_program = to_edge_transform_and_lower(
					torch.export.export(
						model,
						sample_inputs
						if self.conversion[1].quantization != Quantization.float16
						else (sample_inputs[0].half(),),
						dynamic_shapes=dynamic_shapes,
					),
					partitioner=[partitioner],
				).to_executorch()

				with open(output_weights_path, "wb") as file:
					executorch_program.write_to_file(file)

				return ConversionResult(
					format=self.conversion[1],
					image_input=image_input,
					path=output_weights_path,
					output=model_metadata.output,
				)

		def vulkan_setup(self):
			from executorch.backends.vulkan.partitioner.vulkan_partitioner import (  # type: ignore
				VulkanPartitioner,
			)
			from executorch.backends.vulkan.quantizer.vulkan_quantizer import (  # type: ignore
				VulkanQuantizer,
				get_symmetric_quantization_config,
			)

			qparams = get_symmetric_quantization_config(is_dynamic=False, weight_bits=8)
			quantizer = VulkanQuantizer()
			quantizer.set_global(qparams)

			return VulkanPartitioner(), quantizer

		def xnnpack_setup(self):
			from executorch.backends.xnnpack.partition.xnnpack_partitioner import (  # type: ignore
				XnnpackPartitioner,
			)
			from executorch.backends.xnnpack.quantizer.xnnpack_quantizer import (  # type: ignore
				XNNPACKQuantizer,
				get_symmetric_quantization_config,
			)

			qparams = get_symmetric_quantization_config(is_per_channel=True)
			quantizer = XNNPACKQuantizer()
			quantizer.set_global(qparams)

			return XnnpackPartitioner(), quantizer

		def qnn_setup(self):
			from executorch.backends.qualcomm.quantizer.quantizer import (  # type: ignore
				QnnQuantizer,
			)

			return None, QnnQuantizer()

		def convert_qnn(
			self,
			model: "torch.nn.Module",
			sample_inputs: tuple["torch.Tensor"],
			output_weights_path: Path,
		):
			from executorch.backends.qualcomm.utils.utils import (  # type: ignore
				generate_htp_compiler_spec,
				generate_qnn_executorch_compiler_spec,
				get_soc_to_chipset_map,
				to_edge_transform_and_lower_to_qnn,
			)

			output_model_paths: list[tuple[Path, str]] = []

			soc_str = next(
				x[0]
				for x in Executorch.chip_compatibility_lists
				if x[1] == self.conversion[1].chip_compatibility
			)
			soc = get_soc_to_chipset_map()[soc_str]
			try:
				# QNN Compiler Spec
				compile_spec = generate_qnn_executorch_compiler_spec(
					soc_model=soc,  # Your target SoC
					backend_options=generate_htp_compiler_spec(  # HTP Compiler Configuration
						use_fp16=self.conversion[1].quantization
						== Quantization.float16,
					),
				)
				# Lower to QNN backend
				executorch_program = to_edge_transform_and_lower_to_qnn(
					model,
					sample_inputs,
					compile_spec,
				).to_executorch()

				output_model = output_weights_path / soc.name
				with open(output_model, "wb") as file:
					file.write(executorch_program.buffer)

				output_model_paths.append((output_model, soc.name))
			except Exception as e:
				print(e)

			return output_model_paths


capabilities = [
	Executorch(quantization, hw_acceleration, soc)
	for quantization in Quantization
	for hw_acceleration in HardwareAcceleration
	for soc in (
		(x[1] for x in Executorch.chip_compatibility_lists)
		if hw_acceleration == HardwareAcceleration.QNN
		else [None]
	)
]



from pathlib import Path

from ai_parade._toolkit.enums import ExactModelFormat, format_suffixes
from ai_parade._toolkit.runtime_metadata import ModelAttributes, format_distinguishes


def get_weights_name(format: ExactModelFormat) -> Path:
	"""
	Get path to model weights, this path may not exist

	Example:
	>>> get_weights_name(ExactModelFormat(ModelFormat.ONNX, Quantization.none))
	'onnx.all.all.onnx'

	>>> get_weights_name(ExactModelFormat(ModelFormat.ExecutorTorch, Quantization.int8, HardwareAcceleration.XNNPACK))
	'executorch.int8.xnnpack.pte'
	"""
	if ModelAttributes.HWAcceleration in format_distinguishes[format.format]:
		assert format.hw_acceleration_restriction is not None
		hw_acceleration = format.hw_acceleration_restriction.value
	else:
		hw_acceleration = "all"
	name = ".".join(
		[
			format.format.value,
			(
				format.quantization.value
				if ModelAttributes.Quantization in format_distinguishes[format.format]
				else "all"
			),
			hw_acceleration,
		]
	)
	return Path(name + format_suffixes[format.format])



from typing import override

from ai_parade._toolkit.enums import ModelFormat
from ai_parade._toolkit.run.ModelRunner import ModelRunner
from ai_parade._toolkit.run.ONNX import ONNXRunner

from .converter import ConversionResult, Converter, FormatConverter


class ONNX_TensorFlowConverter(FormatConverter):
	"""
	https://github.com/onnx/onnx-tensorflow/tree/main
	"""

	@property
	@override
	def format_conversion(self):
		return (ModelFormat.ONNX, ModelFormat.SavedModel)

	@override
	def create(self, options: Converter.Options):
		super().create(options)
		return self.convert

	def convert(self, runner: ModelRunner):
		assert isinstance(runner, ONNXRunner)
		from onnx_tf.backend import prepare  # type: ignore

		tf_rep = prepare(runner.model)  # prepare tf representation
		output_model_path = self.get_output_weights_path(runner, None)
		tf_rep.export_graph(output_model_path)  # export the model

		return ConversionResult(
			format=self.conversion[1],
			image_input=runner.model_metadata.image_input,
			path=output_model_path,
			output=runner.model_metadata.output,
		)



from pathlib import Path

import pytest
from typer.testing import CliRunner

from ai_parade._cli.ai_parade.main import app

from ._common import setup_cwd_any

runner = CliRunner()


class TestInference:
	repo_path = Path("fixtures/FooBarModel")

	@pytest.fixture(autouse=True)
	def copy_repo(self, tmp_path, monkeypatch):
		setup_cwd_any(
			tmp_path,
			monkeypatch,
			repo_path=self.repo_path,
			skip=[],
		)

	@pytest.mark.parametrize(
		[
			"args",
		],
		[
			pytest.param(*p, id=" ".join(p[0].split()[1:-1]))
			for p in [
				("inference weights.pth FooBarModel --separate-env",),
				("inference models/onnx.float32.all.onnx.json --separate-env",),
				("inference models/onnx.float16.all.onnx.json --separate-env",),
				("inference models/onnx.int8.all.onnx.json --separate-env",),
			]
		],
	)
	def test_inference(self, args):
		result = runner.invoke(app, [*args.split(" ")])

		assert result.exit_code == 0

	@pytest.mark.parametrize(
		[
			"args",
		],
		[
			pytest.param(*p, id=" ".join(p[0].split()[1:-3]))
			for p in [
				("inference weights.pth FooBarModel -o output.json",),
				(
					"inference models/onnx.float32.all.onnx.json -o output.json --separate-env",
				),
			]
		],
	)
	def test_inference_output(self, args):
		result = runner.invoke(app, [*args.split(" ")])

		assert result.exit_code == 0

		assert Path("output.json").exists()



import filecmp
from pathlib import Path


def setup_cwd_any(tmp_path: Path, monkeypatch, repo_path: Path, skip: list[str] = []):
	skip = skip + ["build"]
	for item in (Path(__file__).parent.parent / repo_path).iterdir():
		if item.name in skip:
			continue
		(tmp_path / item.name).symlink_to(item.resolve())

	monkeypatch.chdir(tmp_path)
	monkeypatch.setenv("VIRTUAL_ENV", str(tmp_path / ".venv"))


def text_files_equal(expected: Path, actual: Path) -> None:
	"""Assert that text files at expected and actual paths are identical."""
	from difflib import unified_diff

	with open(expected, "r") as expected_file, open(actual, "r") as actual_file:
		diff = list(
			unified_diff(
				expected_file.readlines(),
				actual_file.readlines(),
				fromfile=str(expected),
				tofile=str(actual),
			)
		)
		assert diff == [], "Unexpected file contents:\n" + "".join(diff)


def files_equal(expected: Path, actual: Path) -> None:
	"""Assert that files/directories at expected and actual paths are identical."""
	if expected.is_file():
		assert filecmp.cmp(expected, actual, shallow=False)

	if expected.is_dir():
		for child in sorted(expected.iterdir()):
			assert child.is_file(), f"Nested directories are not supported: {child}"
			assert filecmp.cmp(child, actual / child.name, shallow=False), (
				f"Artifact mismatch for {expected} vs {actual}"
			)



import logging
from pathlib import Path
from typing import Any, Callable, Coroutine, cast, final, overload

from ai_parade._toolkit.conversion_error import ConversionError
from ai_parade._toolkit.converters.converter import ConversionResult, Converter
from ai_parade._toolkit.converters.select import get_converter
from ai_parade._toolkit.datasets.Dataset import Dataset
from ai_parade._toolkit.enums.formats import ExactModelFormat
from ai_parade._toolkit.get_weights import get_weights_name
from ai_parade._toolkit.metadata.models.final import ModelMetadata
from ai_parade._toolkit.run.ModelRunner import LoadedModelRunner
from ai_parade._toolkit.run.RemoteRunner import RemoteError
from ai_parade._toolkit.run.runner_selector import get_runner
from ai_parade._toolkit.verify import compute_statistics

logger = logging.getLogger(__name__)


class Convert:
	"""
	Extensible converter that can convert between any two formats using available converters.
	Use by calling the instance (it has defined `__call__()`).

	## Path extension points:
	- `_get_result_path_impl`: gets the path to store conversion result JSON

	## Extension points:
	- `on_start`: called as first thing ever.
	- other extension points on each conversion (in other words even on conversions into intermediate formats):
	- `before_conversion`: called before each conversion step
	- `on_loaded_runner`: called when a runner is loaded, still before conversion is done
	- `save_converted_model`: saves converted model to appropriate location, if needed
	- `save_conversion_results`: saves multiple conversion results,
		should not be needed, internally calls `save_converted_model`


	"""

	def __init__(
		self,
		source_format: ExactModelFormat,
		target_format: ExactModelFormat,
		*,
		calibration_dataset: Dataset | None,
		verify_dataset: Dataset | None,
		override: bool,
		use_remote: bool,
		error_immediately: bool,
		metadata: ModelMetadata,
		output: Path | None = None,
		artifacts: Path | None = None,
		default_model: Path | None = None,
	):
		self.source_format = source_format
		self.target_format = target_format
		self.calibration_dataset = calibration_dataset
		self.verify_dataset = verify_dataset
		self.override = override
		self.use_remote = use_remote
		self.error_immediately = error_immediately
		self.metadata = metadata
		self.output = output.resolve() if output is not None else None
		self.artifacts = artifacts.resolve() if artifacts is not None else None

		if default_model is None:
			assert self.output is not None
			self.default_model = self.output / get_weights_name(source_format)
		else:
			self.default_model = default_model

	async def on_start(self, model_path: Path):
		"""
		Called when the converter starts. Before anything else. This method is always called once exactly.

		Arguments:
			model_path: Path to the source model, either weights or conversion result JSON.
		"""
		pass

	async def before_conversion(self):
		"""
		Called before each conversion step.
		"""
		pass

	async def on_loaded_runner(self, runner: LoadedModelRunner):
		"""
		To set up runner.
		Called when a runner is loaded, before each conversion begins. After before_conversion.

		Run the super implementation to set the output directory.
		"""
		assert self.artifacts is not None
		await runner.set_output_directory(self.artifacts)

	@overload
	def get_result_path(self, format_or_conversion_result: ExactModelFormat) -> Path:
		"""
		Gets the path to the conversion result JSON for the given format and hardware acceleration.
		"""
		pass

	@overload
	def get_result_path(self, format_or_conversion_result: ConversionResult) -> Path:
		"""
		Gets the path to the conversion result JSON for the given conversion result.
		"""
		pass

	@final
	def get_result_path(
		self, format_or_conversion_result: ExactModelFormat | ConversionResult
	):
		if isinstance(format_or_conversion_result, ConversionResult):
			format = format_or_conversion_result.format
		else:
			format = format_or_conversion_result
		return self._get_result_path_impl(format)

	def _get_result_path_impl(self, format: ExactModelFormat) -> Path:
		"""
		Gets the path to the conversion result JSON for the given format and hardware acceleration.
		"""

		assert self.output is not None
		model_path = self.output / get_weights_name(format)

		return model_path.with_name(model_path.name + ".json")

	def save_converted_model(self, conversion_result: ConversionResult) -> Path | None:
		"""
		Saves the converted model to the appropriate location and returns the new model path.

		If nothing is done, the model remains at its current path (in artifacts directory).
		The returned path is used in conversion results.
		"""
		new_model_path = self.get_result_path(conversion_result).with_suffix("")
		conversion_result.path.rename(new_model_path)
		return new_model_path

	async def save_failed_conversion_result(self):
		"""
		Saves the failed conversion result to the appropriate location.
		"""
		conversion_result = ConversionResult(
			status="failed",
			format=self.target_format,
			image_input=self.metadata.image_input,
			path=Path(),
			output="",
		)

		with open(self.get_result_path(conversion_result), "w", encoding="utf-8") as f:
			f.write(conversion_result.model_dump_json(indent=2))

	async def save_conversion_results(
		self,
		conversion_results: list[ConversionResult],
		current_target_format: ExactModelFormat | None,
	):
		"""
		Saves multiple conversion results and returns the one matching the current target format and hardware acceleration.
		"""
		requested_result = None
		for conversion_result in conversion_results:
			assert conversion_result.path.is_absolute()
			assert conversion_result.path.exists()
			new_model_path = (
				self.save_converted_model(conversion_result) or conversion_result.path
			)

			# update conversion result path
			result_path = self.get_result_path(conversion_result)
			conversion_result.path = (new_model_path).relative_to(
				result_path.parent, walk_up=True
			)

			# save conversion result
			with open(result_path, "w", encoding="utf-8") as f:
				f.write(conversion_result.model_dump_json(indent=2))
			logger.debug(
				f"Successfully converted model to {conversion_result.format} (at {new_model_path})"
			)

			# check if matches requested format
			if conversion_result.format == current_target_format:
				requested_result = conversion_result
		assert requested_result is not None
		return requested_result

	async def __call__(self, model_path: Path | None = None):
		model_path = model_path or self.default_model
		logger.info(
			f"Going to convert model `{self.metadata.name}` to {self.target_format}"
		)
		await self.on_start(model_path)

		assert model_path.exists(), f"Model path {model_path} does not exist"

		if model_path.suffix == ".json":
			self.metadata, model_path = self.metadata.with_conversion_results(
				model_path
			)

		converter_chains = get_converter(self.source_format, self.target_format)
		logger.debug(f"Converter chains: {converter_chains}")
		if not converter_chains or len(converter_chains) == 0:
			raise ConversionError(
				self.source_format, self.target_format, "Failed to find converter"
			)

		if self.verify_dataset is not None:
			return self.verify(
				lambda: self.__convert__(converter_chains, model_path), model_path
			)
		else:
			await self.__convert__(converter_chains, model_path)

	async def __convert__(
		self, converter_chains: list[list[Converter]], model_path: Path
	) -> ConversionResult | None:
		for converter_chain in converter_chains:
			# reset for each attempt
			current_source_weights_path = model_path
			current_metadata = self.metadata
			requested_result: ConversionResult | None = None
			try:
				for converter in converter_chain:
					# get current conversion variables
					current_source_format, current_target_format = converter.conversion
					current_target_result_path = self.get_result_path(
						current_target_format
					)
					if current_target_result_path.exists() and not self.override:
						# update for next iteration
						current_metadata, current_source_weights_path = (
							current_metadata.with_conversion_results(
								current_target_result_path
							)
						)
						logger.debug(
							f"Model already converted to {current_target_format}. Skipping"
						)
						continue

					# now lets convert
					logger.info(
						f"Converting from {converter.conversion[0]} to {converter.conversion[1]}"
					)
					await self.before_conversion()

					logger.info("Initializing runner")
					async with get_runner(
						current_metadata,
						current_source_weights_path,
						use_remote=self.use_remote,
					) as runner:
						await self.on_loaded_runner(runner)
						calibration_dataset = self.calibration_dataset
						if calibration_dataset is not None:
							calibration_dataset = calibration_dataset.preprocess(
								current_metadata.image_input
							)
						conversion_results = await runner.run(
							converter.create(
								Converter.Options(calibration_data=calibration_dataset)
							)
						)
					if not isinstance(conversion_results, list):
						conversion_results = [conversion_results]

					requested_result = await self.save_conversion_results(
						conversion_results, current_target_format
					)

					# update weights in next iteration
					current_metadata, current_source_weights_path = (
						current_metadata.with_conversion_results(
							self.get_result_path(requested_result), requested_result
						)
					)
				return requested_result
			except MemoryError as e:
				await self.save_failed_conversion_result()
				raise e
			except Exception as e:
				logger.error(
					f"Failed to convert model: {e if type(e) is RemoteError else str(e)}"
				)
				await self.save_failed_conversion_result()
				if self.error_immediately:
					raise e
				break
		raise ConversionError(
			self.source_format,
			self.target_format,
			f"No conversion lead to the target format, tried: {converter_chains}",
		)

	async def verify(
		self,
		convert: Callable[[], Coroutine[Any, Any, ConversionResult | None]],
		model_path: Path | None,
	) -> dict[str, float]:
		"""
		Run inference on both original and converted model and compare results.
		"""
		assert self.artifacts is not None
		assert self.verify_dataset is not None
		async with get_runner(
			self.metadata, model_path or self.default_model, use_remote=self.use_remote
		) as runner:
			await runner.set_output_directory(self.artifacts)
			base_results = await runner.inference(
				self.verify_dataset.preprocess(self.metadata.image_input)
			)

		result = await convert()
		if result is None:
			logger.warning("Conversion is cached, skipping verification")
			return {}

		async with get_runner(
			self.metadata, result.path, use_remote=self.use_remote
		) as runner:
			await runner.set_output_directory(self.artifacts)
			converted_results = await runner.inference(
				self.verify_dataset.preprocess(result.image_input)
			)
		return compute_statistics(cast(list, base_results), converted_results)



import itertools
import random

import pytest

from ai_parade._toolkit.metadata.size import parse_model_size


@pytest.mark.parametrize(
	"expected_order",
	[
		pytest.param(p, id=name)
		for name, p in [
			("Known sizes", [["test", "small"], ["test", "medium"], ["test", "large"]]),
			("Numeric values", [["test", "1"], ["test", "2"], ["test", "3"]]),
			("prefixed numbers", [["test", "b1"], ["test", "b2"], ["test", "b3"]]),
			(
				"Mixed case sizes",
				[["test", "Small"], ["test", "MEDIUM"], ["test", "Large"]],
			),
			(
				"Tricky columns",
				[["large", "small"], ["medium", "medium"], ["small", "large"]],
			),
		]
	],
)
def test_parse_sizes_correct_order(expected_order):
	random.seed(42)
	items = expected_order.copy()
	random.shuffle(items)

	result, unknown = parse_model_size(items, column=1)

	# Sort both lists
	result = sorted(result, key=lambda x: x[1])
	flattened = list(itertools.chain(*[group for key, size, group in result]))
	assert flattened == expected_order
	assert len(unknown) == 0


@pytest.mark.parametrize(
	"items",
	[
		pytest.param(p, id=name)
		for name, p in [
			("Version numbers", [["test", "v1"], ["test", "v2"], ["test", "v3"]]),
			("Invalid input (empty strings)", [["test", ""], ["test", ""]]),
			(
				"numbers, all same",
				[["test", "2", "a"], ["test", "2", "b"], ["test", "2", "c"]],
			),
			(
				"Prefixed numbers, all same",
				[
					["test", "small2", "a"],
					["test", "small2", "b"],
					["test", "small2", "c"],
				],
			),
			(
				"Wrong column",
				[["small", "test"], ["medium", "test"], ["large", "test"]],
			),
			(
				"Just invalid values",
				[["test", "2small"], ["test", "invalid"], ["test", "large2x"]],
			),
		]
	],
)
def test_parse_sizes_unknown_values(items):
	result, unknown = parse_model_size(items, column=1)
	assert len(result) == 0
	assert len(unknown) == len(items)
	assert unknown == items


# todo: test multiple values with same size
# todo: test column/max_column parameters


def test_parse_sizes_with_empty_input():
	items = []
	result, unknown = parse_model_size(items)
	assert len(result) == 0
	assert len(unknown) == 0



import asyncio
import logging
import re
import subprocess
from asyncio.subprocess import Process
from pathlib import Path

from ai_parade._toolkit.constants import TRACE_LVL

_Args = list[str] | list[str | Path] | list[Path]


def _log_command(logger: logging.Logger, args: _Args, **kwargs):
	logger.debug(f"Running command: `{' '.join([str(x) for x in args])}`")
	if kwargs:
		logger.debug(f"\twith kwargs: {kwargs}")


# From https://stackoverflow.com/a/14693789/5739006
_ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")


def better_decode(b: bytes) -> str:
	text = b.decode()
	return _ansi_escape.sub("", text)


def _log_results(
	logger: logging.Logger,
	args: _Args,
	returncode: int,
	stdout: bytes,
	stderr: bytes,
	check: bool,
):
	msg = f"Command: `{' '.join([str(x) for x in args])}` exited with return code {returncode} and output:"
	if stdout is not None:  # type: ignore because it is simply not true
		msg += "\n" + better_decode(stdout)
	if stderr is not None:  # type: ignore because it is simply not true
		msg += "\n" + better_decode(stderr)

	logger.log(logging.ERROR if returncode != 0 else TRACE_LVL, msg)

	if check and returncode != 0:
		raise subprocess.CalledProcessError(returncode, args, stdout, stderr)


def log_subprocess_result(
	result: subprocess.CompletedProcess[bytes],
	*,
	logger: logging.Logger,
	check: bool = True,
):
	_log_results(
		logger, result.args, result.returncode, result.stdout, result.stderr, check
	)
	return result.returncode, better_decode(result.stdout), better_decode(result.stderr)


async def log_subprocess_result_async(
	process: Process,
	args: _Args = [],
	*,
	logger: logging.Logger,
	check: bool = True,
	stdin: bytes | None = None,
):
	stdout, stderr = await process.communicate(input=stdin)
	assert process.returncode is not None
	_log_results(
		logger,
		args,
		process.returncode,
		stdout,
		stderr,
		check,
	)
	return process.returncode, better_decode(stdout), better_decode(stderr)


async def subprocess_run_log_async(
	args: _Args,
	*,
	logger: logging.Logger,
	check: bool = True,
	stdin: str | None = None,
	**kwargs,
):
	_log_command(logger, args, **kwargs)

	process = await asyncio.create_subprocess_exec(
		str(args[0]),
		*map(str, args[1:]),
		stdout=asyncio.subprocess.PIPE,
		stderr=asyncio.subprocess.PIPE,
		stdin=asyncio.subprocess.PIPE if stdin is not None else None,
		**kwargs,
	)
	return await log_subprocess_result_async(
		process,
		args,
		logger=logger,
		check=check,
		stdin=stdin.encode() if stdin is not None else None,
	)


def subprocess_run_log(
	args: _Args, *, logger: logging.Logger, check: bool = True, **kwargs
):
	_log_command(logger, args, **kwargs)
	result = subprocess.run(
		args,
		stdout=asyncio.subprocess.PIPE,
		stderr=asyncio.subprocess.PIPE,
		**kwargs,
	)
	return log_subprocess_result(result, logger=logger, check=check)


def get_ai_parade_logger() -> logging.Logger:
	"""Get the logger for ai_parade"""
	return logging.getLogger("ai_parade._toolkit")



import functools
from typing import Callable, Literal, override

from ai_parade._toolkit.enums import ExactModelFormat, ModelFormat, Quantization
from ai_parade._toolkit.run import ModelRunner
from ai_parade._toolkit.run.TensorFlow import TensorFlowRunner

from .converter import ConversionResult, Converter


class LiteRT_Converter(Converter):
	"""
	Built-in TensorFlow to LiteRT converter, including post training quantization

	PTQ:
	https://docs.pytorch.org/ao/stable/tutorials_source/pt2e_quant_ptq.html
	"""

	def __init__(
		self,
		quantization_type: Literal[Quantization.float16]
		| Literal[Quantization.int8]
		| Literal[Quantization.float32],
	):
		self.quantization_type = quantization_type

	@property
	@override
	def conversion(self):
		return (
			ExactModelFormat(ModelFormat.SavedModel),
			ExactModelFormat(ModelFormat.SavedModel, self.quantization_type),
		)

	@override
	def create(
		self, options: Converter.Options
	) -> Callable[[ModelRunner], ConversionResult]:
		super().create(options)
		assert self.quantization_type not in [Quantization.int8] or options is not None
		return functools.partial(self.convert, options=options)

	def convert(self, runner: ModelRunner, options: Converter.Options):
		assert isinstance(runner, TensorFlowRunner)
		import tensorflow as tf  # type: ignore

		converter = tf.lite.TFLiteConverter.from_saved_model(runner.model)
		match self.quantization_type:
			case Quantization.float32:
				pass
			case Quantization.int8:
				assert options.calibration_data is not None
				# with fallback to other datatypes
				converter.optimizations = [tf.lite.Optimize.DEFAULT]
				converter.representative_dataset = options.calibration_data
			case Quantization.float16:
				converter.optimizations = [tf.lite.Optimize.DEFAULT]
				converter.target_spec.supported_types = [tf.float16]
		tflite_quant_model = converter.convert()

		output_model_path = self.get_output_weights_path(runner, options)
		with open(output_model_path, "wb") as f:
			f.write(tflite_quant_model)

		return ConversionResult(
			format=self.conversion[1],
			image_input=runner.model_metadata.image_input,
			path=output_model_path,
			output=runner.model_metadata.output,
		)



from ..datasets.Dataset import DatasetInfo
from .output import ModelOutput


def postprocess(result: ModelOutput, info: DatasetInfo):
	return bbox_rescale_back(result, info)


def bbox_rescale_back(result: ModelOutput, info: DatasetInfo):
	if "detections" not in result or len(result["detections"]) == 0:
		return result
	assert len(info.source_data_shape) == 3
	assert len(info.data_shape) == 3

	# OpenCV, PyTorch, TF uses (height, width) formats
	source_height, source_width = info.source_data_shape[:2]
	height, width = (
		info.data_shape[:2] if info.data_shape[2] == 3 else info.data_shape[1:3]
	)

	rescale = min(height / source_height, width / source_width)
	x_padding = (width - source_width * rescale) // 2
	y_padding = (height - source_height * rescale) // 2

	for object in result["detections"]:
		if (
			"bbox_xmin_absolute" not in object
			or "bbox_xmax_absolute" not in object
			or "bbox_ymin_absolute" not in object
			or "bbox_ymax_absolute" not in object
		):
			continue
		object["bbox_xmin_absolute"] = int(
			(object["bbox_xmin_absolute"] - x_padding) / rescale
		)
		object["bbox_xmax_absolute"] = int(
			(object["bbox_xmax_absolute"] - x_padding) / rescale
		)
		object["bbox_ymin_absolute"] = int(
			(object["bbox_ymin_absolute"] - y_padding) / rescale
		)
		object["bbox_ymax_absolute"] = int(
			(object["bbox_ymax_absolute"] - y_padding) / rescale
		)

	return result



# backward compatibility <=3.11: typing import TypedDict
from typing_extensions import TypedDict

from ai_parade._toolkit.structural_mapping import MappingDefinition

Output = MappingDefinition


class Classification(TypedDict, total=False):
	label: list[str]

	class_ImageNet1k: int
	class_ImageNet1k_id: str
	class_COCO: int

	class_ImageNet1k_logit: list[float]
	class_COCO_logit: list[float]

	class_ImageNet1k_probability: list[float]
	class_COCO_probability: list[float]


class Detection(Classification, total=False):
	bbox_xmin_absolute: float
	bbox_xmax_absolute: float
	bbox_ymin_absolute: float
	bbox_ymax_absolute: float

	bbox_xcenter_absolute: float
	bbox_ycenter_absolute: float
	bbox_width_absolute: float
	bbox_height_absolute: float

	bbox_xmin_normalized: float
	bbox_xmax_normalized: float
	bbox_ymin_normalized: float
	bbox_ymax_normalized: float

	bbox_xcenter_normalized: float
	bbox_ycenter_normalized: float
	bbox_width_normalized: float
	bbox_height_normalized: float


class ModelOutput(Classification, total=False):
	"""
	Represents the output of the model.

	Currently only vision models are supported.
	"""

	detections: list[Detection]



from enum import StrEnum


class ModelTasks(StrEnum):
	HPE = "pose"
	HPE3D = "3d_pose"
	ObjectTracking = "tracking"
	PersonIdentification = "identification"
	FaceKeypoints = "face"
	DepthEstimation = "depth"
	ObjectDetection = "detection"
	Classification = "classification"



import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, override

if TYPE_CHECKING:
	import onnxruntime as ort  # type: ignore

from .ModelRunnerBase import (
	Dataset,
	ModelMetadata,
	ModelOutput,
	ModelRunnerBase,
)

logger = logging.getLogger(__name__)


class ONNXRunner(ModelRunnerBase["ONNXRunner", "ort.InferenceSession|None"]):
	INPUT_NAME = "input"

	def __init__(self, model_metadata: ModelMetadata, weights: Path) -> None:
		super().__init__(model_metadata, weights)
		self.is_entered = False
		self._model = None

	@override
	async def __aenter__(self):
		try:
			import onnxruntime as ort  # type: ignore

			logger.debug(f"Loading ONNX model from {self.weights}")

			self._model = ort.InferenceSession(
				self.weights,
				providers=["CUDAExecutionProvider", "CPUExecutionProvider"],
			)
		except ImportError:
			logger.warning("ONNX runtime not installed, cannot run inference")
			import onnx  # type: ignore
		self.is_entered = True
		return self

	@override
	async def __aexit__(self, *err: Any):
		self._model = None

	@override
	async def inference(self, dataset: Dataset) -> list[ModelOutput]:
		if self.model is None:
			raise RuntimeError("ONNX runtime not installed, cannot run inference")

		logger.info("Running inference on model")

		outputs = []
		infos = []
		for data, _, data_info in dataset:
			outputs.append(self.model.run(None, {self.INPUT_NAME: data})[0])
			infos.append(data_info)

		return self._postprocess(outputs, infos, self.model_metadata)

	@property
	@override
	def model(self):
		return self._model



import asyncio
import pickle
import socket
import struct
from typing import Any


class ConnectionClosed(Exception):
	"""Exception raised when the connection is closed"""

	def __init__(self, message: str):
		super().__init__(message)
		self.message = message


def send(socket: socket.socket, data: Any):
	pickled_data = pickle.dumps(data, 4)
	data_len = struct.pack(">L", len(pickled_data))
	socket.sendall(data_len + pickled_data)


async def send_async(writer: asyncio.StreamWriter, data: Any):
	pickled_data = pickle.dumps(data, 4)
	data_len = struct.pack(">L", len(pickled_data))
	writer.write(data_len + pickled_data)
	await writer.drain()


BUFF_SIZE = 4096  # 4 KiB


def receive(socket: socket.socket):
	return asyncio.run(receive_async(socket))


async def receive_async(reader: asyncio.StreamReader | socket.socket):
	raw_len = (
		await reader.read(4)
		if isinstance(reader, asyncio.StreamReader)
		else reader.recv(4)
	)
	if len(raw_len) != 4:
		if raw_len == b"":
			raise ConnectionClosed("Empty data received")
		else:
			raise RuntimeError(f"Wrong data received {raw_len} (expected data length)")
	data_len = struct.unpack(">L", raw_len)[0]

	fragments = []
	data_read = 0
	fragment = b""
	while data_read < data_len:
		fragment = (
			await reader.read(BUFF_SIZE)
			if isinstance(reader, asyncio.StreamReader)
			else reader.recv(BUFF_SIZE)
		)
		if len(fragment) == 0:
			raise ConnectionClosed("Empty data received")
		fragments.append(fragment)
		data_read += len(fragment)
	assert len(fragment) < BUFF_SIZE

	return pickle.loads(b"".join(fragments))



import pytest


@pytest.mark.parametrize(
	["raw", "mapping", "expected"],
	[
		pytest.param(*p[1:], id=p[0])
		for p in [
			[
				"flat dictionary",
				{"a": 1, "b": 2},
				{"a": "alpha", "b": "beta"},
				{"alpha": 1, "beta": 2},
			],
			[
				"flat list",
				[1, 2, 3],
				["a", "b", "c"],
				{"a": 1, "b": 2, "c": 3},
			],
			[
				"flat list not matching length",
				[1, 2, 3],
				["a", "b"],
				{"a": 1, "b": 2},
			],
			[
				"flat tuple",
				(1, 2, 3),
				["a", "b", "c"],
				{"a": 1, "b": 2, "c": 3},
			],
			[
				"nested structures",
				{"a": 1, "b": {"c": 2}, "d": [3, 4]},
				{"a": "alpha", "b": {"c": "gamma"}, "d": ["delta", "epsilon"]},
				{"alpha": 1, "gamma": 2, "delta": 3, "epsilon": 4},
			],
			[
				"nested structures with additional values",
				{"a": 1, "b": {"c": 2, "d": 3}, "d": [3, 4, 6], "e": 5},
				{"a": "alpha", "b": {"c": "gamma"}, "d": ["delta", "epsilon"]},
				{"alpha": 1, "gamma": 2, "delta": 3, "epsilon": 4},
			],
		]
	],
)
def test_positive(raw, mapping, expected):
	from ai_parade._toolkit.structural_mapping import structural_map

	result = structural_map(raw, mapping)
	assert result == expected


@pytest.mark.parametrize(
	["raw", "mapping", "expected"],
	[
		pytest.param(*p[1:], id=p[0])
		for p in [
			[
				"skipping in a list",
				[1, 2, 3, 4, 5, 6],
				[
					"a",
					"...skip",
					"c",
					"...skip",
					"e",
				],
				{"a": 1, "c": 3, "e": 5},
			],
			[
				"repeating patterns (single value)",
				[1, 2, 3, 4, 5, 6],
				["a", "...repeat"],
				{"a": [1, 2, 3, 4, 5, 6]},
			],
			[
				"repeating patterns (single value, no input)",
				[],
				["a", "...repeat"],
				{"a": []},
			],
			[
				"repeating patterns (single value, 1 repetition)",
				[1],
				["a", "...repeat"],
				{"a": [1]},
			],
			[
				"repeating patterns (multiple values, no input)",
				[],
				["odd", "even", "...repeat"],
				{"odd": [], "even": []},
			],
			[
				"repeating patterns (multiple values, partial input)",
				[1],
				["odd", "even", "...repeat"],
				{"odd": [1], "even": []},
			],
			[
				"repeating patterns (multiple values)",
				[1, 2, 3, 4, 5, 6],
				["odd", "even", "...repeat"],
				{"odd": [1, 3, 5], "even": [2, 4, 6]},
			],
			[
				"repeating patterns (multiple values, 1 repetition)",
				[1, 2],
				["odd", "even", "...repeat"],
				{"odd": [1], "even": [2]},
			],
			[
				"repeating patterns with skipping",
				[1, 2, 3, 4, 5, 6],
				["odd", "even", "...skip", "...repeat"],
				{"odd": [1, 4], "even": [2, 5]},
			],
			[
				"named repeating patterns (no input)",
				[],
				["a", "b", "...skip", "...repeat as group"],
				{"group": []},
			],
			[
				"named repeating patterns",
				[1, 2, 3, 4, 5, 6],
				["a", "b", "...skip", "...repeat as group"],
				{"group": [{"a": 1, "b": 2}, {"a": 4, "b": 5}]},
			],
			[
				"named repeating patterns (merge)",
				{"a_values": [1, 2], "b_values": [3, 4]},
				{
					"a_values": ["a", "...repeat as group"],
					"b_values": ["b", "...repeat as group"],
				},
				{"group": [{"a": 1, "b": 3}, {"a": 2, "b": 4}]},
			],
			[
				"nested named repeating patterns",
				[{"a": [1, 2, 3, 4, 5, 6]}, {"a": [7, 8, 9, 10, 11, 12]}],
				[
					{"a": ["alpha", "beta", "...skip", "...repeat as inner_group"]},
					"...repeat as outer_group",
				],
				{
					"outer_group": [
						{
							"inner_group": [
								{"alpha": 1, "beta": 2},
								{"alpha": 4, "beta": 5},
							]
						},
						{
							"inner_group": [
								{"alpha": 7, "beta": 8},
								{"alpha": 10, "beta": 11},
							]
						},
					]
				},
			],
		]
	],
)
def test_special_symbols(raw, mapping, expected):
	from ai_parade._toolkit.structural_mapping import structural_map

	result = structural_map(raw, mapping)
	assert result == expected


@pytest.mark.parametrize(
	["raw", "mapping"],
	[
		pytest.param(*p[1:], id=p[0])
		for p in [
			[
				"missing dict key",
				{"a": 1},
				{"a": "alpha", "b": "beta"},
			],
			[
				"missing tuple index",
				[1, 2],
				["a", "b", "c"],
			],
			[
				"missing nested dict key",
				{"a": 1, "b": {"c": 2}},
				{"a": "alpha", "b": {"c": "gamma", "d": "delta"}},
			],
			[
				"missing nested tuple index",
				{"a": [1, 2], "b": 3},
				{"a": ["alpha", "beta", "gamma"], "b": "delta"},
			],
			[
				"empty dict with mapping",
				{},
				{"a": "alpha", "b": "beta"},
			],
			[
				"empty list with tuple mapping",
				[],
				["a", "b"],
			],
			[
				"nested repeating patterns",
				[{"group": [1, 2, 3, 4, 5, 6]}, {"group": [7, 8, 9, 10, 11, 12]}],
				[{"group": ["odd", "even", "...repeat"]}, "...repeat"],
			],
			[
				"nested repeating patterns (inner named)",
				[{"group": [1, 2, 3, 4, 5, 6]}, {"group": [7, 8, 9, 10, 11, 12]}],
				[{"group": ["a", "b", "...skip", "...repeat as group"]}, "...repeat"],
			],
			[
				"named repeating patterns - input not divisible",
				[1, 2],
				["a", "b", "...skip", "...repeat as group"],
			],
		]
	],
)
def test_negative_missing_symbols(raw, mapping):
	from ai_parade._toolkit.structural_mapping import (
		StructuralMappingError,
		structural_map,
	)

	with pytest.raises(StructuralMappingError):
		structural_map(raw, mapping)



from abc import ABC
from pathlib import Path
from typing import (
	Any,
	Callable,
	Iterable,
	cast,
	override,
)

import cv2 as cv
import numpy as np

from ai_parade._toolkit.data.output import ModelOutput

from .Dataset import Dataset, DatasetInfo, Output_unbatched


class FileDataset(Dataset[tuple[Path, dict]], ABC):
	"""Dataset from files.
	Loads and parses image data.
	The source of the data is provided from derived classes by implementing walk().
	"""

	def __init__(
		self,
		annotation_mapping: Callable[[Any], ModelOutput],
	) -> None:
		super().__init__()
		self._annotation_mapping = annotation_mapping

	@override
	def parse_data(
		self, iterable: Iterable[tuple[Path, dict]], should_load: bool
	) -> Iterable[Output_unbatched]:
		for data, annotation in iterable:
			if should_load:
				image = cv.imread(str(data))
			else:
				image = np.array([])
			annotation = cast(ModelOutput, self._annotation_mapping(annotation))

			info = DatasetInfo(
				source_data_shape=image.shape,
				data_source=str(data),
				data_shape=image.shape,
			)
			yield image, annotation, info



import logging
import os
from pathlib import Path
from typing import Any, Callable, override

import xmltodict

from ai_parade._toolkit.data.output import ModelOutput

from .Dataset import DatasetSplit
from .FileDataset import FileDataset

logger = logging.getLogger(__name__)


class PascalVOC(FileDataset):
	def __init__(
		self,
		dataset_path: Path,
		split: DatasetSplit,
		annotation_mapping: Callable[[Any], ModelOutput],
		annotations_dir: str = "Annotations",
		data_dir: str = "Data",
	):
		super().__init__(annotation_mapping)
		self._dataset_path = dataset_path
		match split:
			case DatasetSplit.train:
				split_dir = "train"
			case DatasetSplit.val:
				split_dir = "val"
			case DatasetSplit.test:
				split_dir = "test"

		# handle cases like: ILSVRC/Data/VID when there are no other datasets (CLS, DET, VID, etc.)
		data_content = list((dataset_path / data_dir).iterdir())
		if len(data_content) == 1:
			self._data_path = data_content[0] / split_dir
		else:
			self._data_path = dataset_path / data_dir / split_dir

		annotations_content = list((dataset_path / annotations_dir).iterdir())
		if len(annotations_content) == 1:
			self._annotations_path = annotations_content[0] / split_dir
		else:
			self._annotations_path = dataset_path / annotations_dir / split_dir

		if not self._data_path.exists():
			raise ValueError(f"Data directory {self._data_path} does not exist")
		if not self._annotations_path.exists():
			raise ValueError(
				f"Annotations directory {self._annotations_path} does not exist"
			)

	@override
	def walk(self):
		"""Walk the dataset directory tree.

		Returns:
			 iterator of tuples of (path to the annotation file and path to the data file)
		"""
		""" Example paths:
		- Data/VID/test/ILSVRC2015_test_00000000/000000.JPEG
		- Data/VID/train/ILSVRC2015_VID_train_0000/ILSVRC2015_train_00000000/000000.JPEG
		- Annotations/VID/train/ILSVRC2015_VID_train_0000/ILSVRC2015_train_00000000/000000.xml
		"""
		data_path_len = len(self._data_path.parts)
		for path, dirs, files in os.walk(self._data_path):
			# this makes the walking of the directories in sorted order
			# https://stackoverflow.com/a/18282602/5739006
			dirs.sort()
			for file in sorted(files):
				data_path = Path(path) / file
				# this is basically data_path.relative_to(self._data_path)
				# but it runs much faster
				relative_data_path = Path(*data_path.parts[data_path_len:])
				annotation_path = (
					self._annotations_path / relative_data_path.with_suffix(".xml")
				)
				if annotation_path.exists():
					with open(annotation_path) as fd:
						annotation = xmltodict.parse(fd.read())
				else:
					logger.warning(f"Annotation file {annotation_path} does not exist")
					annotation = {}
				yield data_path, annotation



import sys
from pathlib import Path

from ai_parade._toolkit.constants import get_site_packages_path


class SysPathCheckpoint:
	def __init__(self, temp_repository: Path | None = None):
		self.temp_repository_path = temp_repository
		self._sys_path = []

	def __enter__(self):
		self._sys_path = sys.path.copy()
		if self.temp_repository_path is not None:
			try:
				site_packages = get_site_packages_path(self.temp_repository_path)
				if site_packages.exists():
					sys.path.insert(0, str(site_packages))
			except StopIteration:
				pass
			# this give precedence to repository over the site packages
			sys.path.insert(0, str(self.temp_repository_path))
		return self

	def __exit__(self, exc_type, exc_value, traceback):
		sys.path = self._sys_path



from dataclasses import dataclass
from enum import StrEnum

from .AIRuntime import AIRuntime
from .hw_acceleration import HardwareAcceleration


class ModelFormat(StrEnum):
	PyTorch = "pytorch"
	ExecutorTorch = "executorch"
	SavedModel = "savedmodel"
	LiteRT = "tflite"
	ONNX = "onnx"
	SNPE = "snpe"
	NCNN = "ncnn"
	MNN = "mnn"
	bolt = "bolt"
	MACE = "mace"
	PaddleLite = "paddlelite"
	# New AI Runtime: update here


class Quantization(StrEnum):
	none = "float32"
	float32 = "float32"
	float16 = "float16"
	int8 = "int8"
	int4 = "int4"


@dataclass(frozen=True)
class ExactModelFormat:
	"""
	Represents a exact model format.
	"""

	format: ModelFormat
	"""
	Model format
	"""
	quantization: Quantization = Quantization.none
	"""
	Quantization of the model
	"""
	hw_acceleration_restriction: HardwareAcceleration | None = None
	"""
	Hardware acceleration of the converted model
	"""
	chip_compatibility: tuple[str, ...] | None = None
	"""
	Chip compatibility of the converted model
	"""


formats_preferred_runtime: dict[ModelFormat, AIRuntime] = {
	ModelFormat.LiteRT: AIRuntime.LiteRT,
	ModelFormat.NCNN: AIRuntime.ncnn,
	ModelFormat.ONNX: AIRuntime.ONNXRuntime,
	ModelFormat.PyTorch: AIRuntime.PyTorch,
	ModelFormat.SavedModel: AIRuntime.TensorFlow,
}

format_suffixes: dict[ModelFormat, str] = {
	ModelFormat.SavedModel: ".savedmodel",
	ModelFormat.PyTorch: ".pth",
	ModelFormat.SavedModel: ".pb",
	ModelFormat.LiteRT: ".tflite",
	ModelFormat.ONNX: ".onnx",
	# this should be directory with ncnn.param and ncnn.bin
	ModelFormat.NCNN: ".ncnn",
	ModelFormat.ExecutorTorch: ".pte",
	# New AI Runtime: update here
}



import logging
from typing import TYPE_CHECKING, Any, override

if TYPE_CHECKING:
	import tensorflow as tf  # type: ignore

from .ModelRunnerBase import (
	Dataset,
	ModelOutput,
	ModelRunnerBase,
	TReturn,
)

logger = logging.getLogger(__name__)


class TensorFlowRunner(ModelRunnerBase["TensorFlowRunner", "tf.Graph"]):
	@override
	async def __aenter__(self):
		logger.debug(f"Loading TensorFlow model from {self.weights}")
		raise NotImplementedError
		return self

	@override
	async def __aexit__(self, *err: Any):
		self._model = None

	@override
	async def inference(self, dataset: Dataset) -> list[ModelOutput]:
		logger.info("Running inference on model")
		raise NotImplementedError

		# return self._postprocess(outputs, infos, self.model_metadata)



import io
import logging
import sys
import traceback
import warnings
from enum import Enum
from pathlib import Path
from types import TracebackType
from typing import Any, override


class CaptureOutputs:
	"""Capture the outputs of a process, including stdout, stderr, and logs.

	What is captured:
	- stdout
	- stderr
	- logs
	- warnings
	- errors/exceptions

	Once exited, the class contains the captured outputs in its public attributes.
	"""

	def __init__(
		self,
		save_dir: Path | None = None,
		single_output: bool = True,
		catch_errors: bool = True,
		**kwargs: Any,
	):
		"""
		Args:
			save_dir: Directory to save the captured outputs.
			single_output: If `True` all outputs will be saved to the logs field / file.
			catch_errors: If `True` all exceptions will be caught and saved to `error` field.
		"""
		# path attributes
		self.save_dir = save_dir
		"""Directory to save the captured outputs."""

		# setup attributes
		self._catch_errors = catch_errors
		self._single_output = single_output

		# Public attributes for inspecting the outputs
		self.error = ""
		"""Error messages of error that cause the CaptureOutputs to exit."""
		self.stdout = ""
		"""Captured stdout. Empty if `mixed_stdout_logger` is set to `True`."""
		self.stderr = ""
		"""Captured stderr. Empty if `mixed_stdout_stderr` is set to `True`."""
		self.logs = ""
		"""Captured logs"""
		self.warnings = []
		"""Captured warning messages."""
		self._warnings = []

		# logging, stdout, and stderr streams
		self._log_stream = io.StringIO()
		self._stdout_stream = self._log_stream if single_output else io.StringIO()
		self._stderr_stream = self._log_stream if single_output else io.StringIO()

		# capture handlers
		self._handlers_original = {}
		self._handler = None
		self._warnings_catcher = None

		self._is_entered = False

	def __enter__(self):
		self._is_entered = True

		# capture logs
		# - set up new log handler
		self._handler = logging.StreamHandler(self._log_stream)
		self._handler.setFormatter(
			logging.Formatter("[%(levelname)-9s%(asctime)s][%(name)s] %(message)s")
		)
		self._handler.addFilter(
			lambda record: not hasattr(record.msg, "__len__") or len(record.msg) != 0
		)
		self._handler.setLevel(0)

		# - save original handlers and replace them
		self._handlers_original: dict[str | None, list[logging.Handler]] = {
			None: logging.root.handlers
		}
		logging.root.handlers = [self._handler]
		for key, value in logging.root.manager.loggerDict.items():
			if isinstance(value, logging.Logger):
				self._handlers_original[key] = value.handlers
				value.handlers = [] if value.propagate else [self._handler]

		# capture stdout and stderr
		# note we can't use the sys.__stderr__ and sys.__stdout__
		# because it breaks notebooks i.e. the notebooks redirects the streams
		self._original_stdout = sys.stdout
		self._original_stderr = sys.stderr
		sys.stdout = self._stdout_stream
		sys.stderr = self._stderr_stream

		# capture warnings
		if self._single_output:
			self._logging_captures_warnings = logging._warnings_showwarning is not None  # type: ignore
			logging.captureWarnings(True)
		else:
			self._warnings_catcher = warnings.catch_warnings(record=True)
			self._warnings = self._warnings_catcher.__enter__()

		return self

	def __exit__(
		self,
		error_type: type[BaseException] | None,
		error: BaseException | None,
		tb: TracebackType | None,
	):
		self._is_entered = False

		# save to fields and file
		self._save_to_fields(error_type, error, tb)
		if self.save_dir is not None:
			self.save(reset=False)

		# restore stdout and stderr
		sys.stderr = self._original_stderr
		sys.stdout = self._original_stdout

		# restore warnings
		if self._single_output:
			logging.captureWarnings(self._logging_captures_warnings)
		else:
			assert self._warnings_catcher is not None
			self._warnings_catcher.__exit__(error_type, error, tb)

		# restore logging
		assert self._handler is not None
		logging.root.handlers = self._handlers_original.pop(None)
		for key, value in self._handlers_original.items():
			assert key is not None
			logger = logging.root.manager.loggerDict[key]
			assert isinstance(logger, logging.Logger)
			logger.handlers = value

		# close streams
		self._log_stream.close()
		self._stdout_stream.close()
		self._stderr_stream.close()

		return self._catch_errors

	def _save_to_fields(
		self,
		error_type: type[BaseException] | None = None,
		error: BaseException | None = None,
		tb: TracebackType | None = None,
	):
		"""Save the captured outputs to the public fields."""
		self.logs += self._log_stream.getvalue()
		self._log_stream = io.StringIO()

		if not self._single_output:
			self.stdout += self._stdout_stream.getvalue()
			self._stdout_stream = io.StringIO()

			self.stderr += self._stderr_stream.getvalue()
			self._stderr_stream = io.StringIO()

			if error_type is not None:
				self.error = "".join(traceback.format_exception(error_type, error, tb))
		else:
			if error_type is not None:
				self.logs += "".join(traceback.format_exception(error_type, error, tb))

		self.warnings += self._warnings

	def reset(self):
		"""Reset the captured outputs."""
		self.error = ""
		self.stdout = ""
		self.stderr = ""
		self.logs = ""
		self.warnings = []

	def update(self, other: "CaptureOutputs"):
		"""Update the captured outputs with another instance."""
		self.error += other.error
		self.stdout += other.stdout
		self.stderr += other.stderr
		self.logs += other.logs
		self.warnings += other.warnings

	def __setstate__(self, state: dict):
		self.__init__(**state)
		self.__dict__.update(state)
		return self

	@override
	def __getstate__(self):
		return {
			key: self.__dict__[key]
			for key in [
				# data:
				"error",
				"stdout",
				"stderr",
				"logs",
				"warnings",
				# state:
				"temp_dir",
				# setup:
				"_with_temp_dir",
				"_single_output",
				"_catch_errors",
			]
		}

	@property
	def summary(self):
		state = LogSummary.State.OK if self.error == "" else LogSummary.State.Error
		return LogSummary(
			state,
			error=[self.error],
			warnings=self._warnings,
			log_dir=[self.save_dir] if self.save_dir is not None else [],
		)

	def save(self, save_dir: Path | None = None, reset: bool = False):
		"""
		Save the captured outputs to the specified directory.

		Args:
			save_dir: Directory to save the captured outputs. Falls back to `self.save_dir` if not provided.
			reset: If `True`, reset the captured outputs after saving.
		"""
		save_dir = save_dir or self.save_dir
		if save_dir is None:
			raise ValueError("save_dir is not set")

		if self._is_entered:
			self._save_to_fields()
			if reset:
				self.reset()

		save_dir.mkdir(parents=True, exist_ok=True)
		for key in ["error", "warnings", "logs", "stdout", "stderr"]:
			assert hasattr(self, key)
			value = getattr(self, key)
			if value is not None and len(value) != 0:
				with open(save_dir / f"{key}.txt", "a") as file:
					if isinstance(value, list):
						value = "\n".join([str(x) for x in value])
					file.write(str(value))


class LogSummary:
	class State(Enum):
		OK = 0
		Error = 1

	def __init__(
		self,
		state: State,
		*,
		error: list[str],
		warnings: list[warnings.WarningMessage],
		log_dir: list[Path],
	):
		self.state = state
		self.error = error
		self.warnings = warnings
		self.log_dir = log_dir

	def merge(self, other: "LogSummary"):
		return LogSummary(
			self.State(max(self.state.value, other.state.value)),
			error=self.error + other.error,
			warnings=self.warnings + other.warnings,
			log_dir=self.log_dir + other.log_dir,
		)



import logging
import shutil
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Literal

from ai_parade._toolkit.logging import subprocess_run_log

logger = logging.getLogger(__name__)


class ONNXSimplify:
	def __init__(
		self,
		simplifier: Literal["onnxslim"] | Literal["onnxsim"] = "onnxslim",
	):
		self.simplifier = simplifier

	def simplify(
		self,
		onnx_path: Path,
	):
		with NamedTemporaryFile() as tmp_file:
			logger.info(f"Running {self.simplifier} on ONNX model")

			returncode, _, _ = subprocess_run_log(
				[
					self.simplifier,
					onnx_path,
					tmp_file.name,
				],
				logger=logger,
			)
			if returncode == 0:
				shutil.move(
					onnx_path, onnx_path.with_stem(onnx_path.stem + "_unsimplified")
				)
				shutil.copy(tmp_file.name, onnx_path)

			return returncode == 0



import os
import numpy as np
import tempfile, zipfile
import torch
import torch.nn as nn
import torch.nn.functional as F
try:
    import torchvision
    import torchaudio
except:
    pass

class Model(nn.Module):
    def __init__(self):
        super(Model, self).__init__()

        self.model_0 = nn.Conv2d(bias=True, dilation=(1,1), groups=1, in_channels=3, kernel_size=(8,8), out_channels=3, padding=(0,0), padding_mode='zeros', stride=(1,1))

        archive = zipfile.ZipFile('build/model.pnnx.bin', 'r')
        self.model_0.bias = self.load_pnnx_bin_as_parameter(archive, 'model.0.bias', (3), 'float32')
        self.model_0.weight = self.load_pnnx_bin_as_parameter(archive, 'model.0.weight', (3,3,8,8), 'float32')
        archive.close()

    def load_pnnx_bin_as_parameter(self, archive, key, shape, dtype, requires_grad=True):
        return nn.Parameter(self.load_pnnx_bin_as_tensor(archive, key, shape, dtype), requires_grad)

    def load_pnnx_bin_as_tensor(self, archive, key, shape, dtype):
        fd, tmppath = tempfile.mkstemp()
        with os.fdopen(fd, 'wb') as tmpf, archive.open(key) as keyfile:
            tmpf.write(keyfile.read())
        m = np.memmap(tmppath, dtype=dtype, mode='r', shape=shape).copy()
        os.remove(tmppath)
        return torch.from_numpy(m)

    def forward(self, v_0):
        v_1 = self.model_0(v_0)
        v_2 = torch.flatten(v_1, end_dim=-1, start_dim=1)
        return v_2

def export_torchscript():
    net = Model()
    net.float()
    net.eval()

    torch.manual_seed(0)
    v_0 = torch.rand(1, 3, 8, 8, dtype=torch.float)

    mod = torch.jit.trace(net, v_0)
    mod.save("build/model_pnnx.py.pt")

def export_onnx():
    net = Model()
    net.float()
    net.eval()

    torch.manual_seed(0)
    v_0 = torch.rand(1, 3, 8, 8, dtype=torch.float)

    torch.onnx.export(net, v_0, "build/model_pnnx.py.onnx", export_params=True, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK, opset_version=13, input_names=['in0'], output_names=['out0'])

def export_pnnx():
    net = Model()
    net.float()
    net.eval()

    torch.manual_seed(0)
    v_0 = torch.rand(1, 3, 8, 8, dtype=torch.float)

    import pnnx
    pnnx.export(net, "build/model_pnnx.py.pt", v_0)

def export_ncnn():
    export_pnnx()

@torch.no_grad()
def test_inference():
    net = Model()
    net.float()
    net.eval()

    torch.manual_seed(0)
    v_0 = torch.rand(1, 3, 8, 8, dtype=torch.float)

    return net(v_0)

if __name__ == "__main__":
    print(test_inference())



import cv2
import torch
import torch.nn as nn

SIZE = 8


# Define a simple CNN using Sequential
class ChannelSum(nn.Module):
	def __init__(self):
		super(ChannelSum, self).__init__()
		self.model = nn.Sequential(
			nn.Conv2d(3, 3, kernel_size=SIZE),
			nn.Flatten(),
		)

	def forward(self, x):
		return self.model(x)


def main():
	model = ChannelSum()
	with torch.no_grad():
		for m in model.modules():
			if isinstance(m, nn.Conv2d):
				nn.init.constant_(m.bias, 0)
				m.bias[2] = -(SIZE * SIZE - 1)
				nn.init.constant_(m.weight, 0)
				# m.weight.shape = (out_channels, in_channels, kernel_size, kernel_size)
				# this essentially sums up the channel values
				for i in range(m.weight.shape[0]):
					m.weight[i, i, :, :] = 1

	print("running")
	model.eval()
	blue_only = torch.stack(
		[
			torch.zeros(1, 1, SIZE, SIZE),
			torch.zeros(1, 1, SIZE, SIZE),
			torch.ones(1, 1, SIZE, SIZE),
		]
	).reshape(1, 3, SIZE, SIZE)

	print("Expected results:", model(blue_only))
	print(
		"Channel order mismatch results:",
		model(blue_only.permute([0, 2, 3, 1]).reshape(1, 3, SIZE, SIZE)),
	)

	print("saving")
	torch.save(model.state_dict(), "weights.pth")
	cv2.imwrite(
		"testImage.png",
		blue_only.permute([0, 2, 3, 1]).squeeze().numpy()[:, :, ::-1] * 256,
	)


if __name__ == "__main__":
	main()



import importlib
import pathlib
import pkgutil

from ..enums import AIRuntime, HardwareAcceleration, ModelFormat, Platform, Quantization
from ._common import ModelAttributes

# Discover all modules and subpackages under this package
__all__ = []
submodules = [
	importlib.import_module(f"{__name__}.{module_info.name}")
	for module_info in pkgutil.iter_modules([str(pathlib.Path(__file__).parent)])
	if module_info.name != "_common"
]

capabilities: dict[
	AIRuntime,
	dict[Platform, dict[Quantization, set[HardwareAcceleration]]],
] = {module.runtime: module.capabilities for module in submodules}
"""
Capabilities of each AI runtime on each platform for each quantization and hardware acceleration.
"""

format_distinguishes: dict[ModelFormat, list[ModelAttributes]] = {
	module.format: module.format_distinguishes for module in submodules
}
"""
A dictionary mapping model formats to lists of attributes that the format distinguishes.
"""

format: dict[AIRuntime, ModelFormat] = {
	module.runtime: module.format for module in submodules
}
"""
A dictionary mapping AI runtimes to model formats.
"""

__all__ = ["capabilities", "format_distinguishes", "format", "ModelAttributes"]



import re
from itertools import groupby

sizes = [
	# small metric
	"yocto",
	"zepto",
	"atto",
	"femto",
	"pico",
	"nano",
	"micro",
	"milli",
	# small whatever
	"tiny",
	"mini",
	# x smalls
	"xxxs",
	"xxxsmall",
	"xxs",
	"xxsmall",
	"xs",
	"xsmall",
	"s",
	"small",
	# base
	"b",
	"base",
	"m",
	"medium",
	# x larges
	"l",
	"large",
	"x",
	"xl",
	"xlarge",
	"xxl",
	"xxlarge",
	"xxxl",
	"xxxlarge",
	# large whatever
	"huge",
	"giant",
	"gigantic",
]


def parse_model_size(
	items: list[list[str]], column: int = 0, max_column: int = 1
) -> tuple[list[tuple[list[str], float | int, list[list[str]]]], list[list[str]]]:
	"""
	Try to parse splitted model names to find the model size.
	Looks only on columns (indexes) between column <= here < max_column

	It tests columns against this heuristics to find the size, first found is used:
		- if column is in the size list (case insensitive), use its index as a size
		- if column doesn't have the same value for all items:
			- if column is numeric, use it as a size
			- if column has a common prefix and then a number, use it as a size

	See tests for examples.

	Args:
		items: list of models each model is a lists of strings
		column: first column (index) to try to extract the size from
		max_column: last column (index) to try to extract the size from

	Returns:
		list of tuples of size (common prefix, size, models) and list of models with unknown size
	"""
	sizes_map: list[tuple[list[str], float | int, list[list[str]]]] = []
	unknown_list: list[list[str]] = []
	maybe_known: list[tuple[tuple[str, str], list[str]]] = []

	# lets work only on prefixes:
	for k, group in groupby(items, lambda x: x[column] if len(x) > column else ""):
		size = None
		group = list(group)
		name = group[0][:column]
		try:
			# no numbers: try to find it in sizes
			if k.isalpha():
				size = sizes.index(k.lower())
			# if not same for all items
			elif len(group) != len(items):
				# its number: use it
				if k.isnumeric():
					size = float(k)
				else:
					# it ends with a number, we will see if other groups also ends with a number (different), if so we will use it
					result = re.search(r"\d+$", k)
					if result is not None:
						index = result.start()
						if k[:index] != "v":
							maybe_known.extend(
								[((k[:index], k[index:]), item) for item in group]
							)
							continue
			if size is not None:
				sizes_map.append((name, size, group))
		except ValueError:
			pass
		if size is None:
			# lets try next column if possible
			if len(group) > 1 and column + 1 < max_column:
				parsed, unknown = parse_model_size(group, column + 1)
				sizes_map += parsed
				unknown_list += unknown
			else:
				unknown_list += group

	# here we regroup the items that ends with a number, first group them by the alpha part
	for key, group in groupby(maybe_known, lambda x: x[0][0]):
		group = list(group)
		name = group[0][1][:column] + [key]
		# here we group them by the number part
		for key, number_group in groupby(group, lambda x: x[0][1]):
			number_group = list(number_group)
			number_group_items = [item[1] for item in number_group]

			# if there are multiple different numbers
			if len(number_group) != len(group):
				sizes_map.append((name, float(key), number_group_items))
			else:
				unknown_list += number_group_items

	return sizes_map, unknown_list



import logging
from typing import Any, Callable, Generic, Sequence, override

from ai_parade._toolkit.data.output import ModelOutput
from ai_parade._toolkit.data.postprocessing import postprocess
from ai_parade._toolkit.datasets.Dataset import Dataset as Dataset
from ai_parade._toolkit.datasets.Dataset import DatasetInfo
from ai_parade._toolkit.metadata.models.final import ModelMetadata
from ai_parade._toolkit.run.ModelRunner import (  # type: ignore
	LoadedModelRunner,
	ModelRunner,
	TChild,
	TModel,
	TReturn,
)
from ai_parade._toolkit.structural_mapping import structural_map

logger = logging.getLogger(__name__)


# backward compatibility <=3.11: generic class notation
class ModelRunnerBase(
	Generic[TChild, TModel],
	ModelRunner[TChild, TModel],
	LoadedModelRunner[TChild, TModel],
):
	def _postprocess(
		self,
		outputs: list[Any],
		infos: list[Sequence[DatasetInfo]],
		model_metadata: ModelMetadata,
	):
		results: list[ModelOutput] = []
		for output_batch, info_batch in zip(outputs, infos):
			for output, model_info in zip(output_batch, info_batch):
				mapped = self._output_mapping(output, model_metadata)
				results.append(postprocess(mapped, model_info))

		return results

	def _output_mapping(
		self, output: Any, model_metadata: ModelMetadata
	) -> ModelOutput:
		# ?todo: use pydantic to check types? ... only in debug mode
		return structural_map(output, model_metadata.output)

	@override
	async def run(self, what: Callable[[TChild], TReturn]) -> TReturn:
		logger.debug("Running %s", what)
		return what(self)  # type: ignore



from src import ModelFormat
from src.data.imageInput import ImageInput


def get_DL_Lib_preset(dl_lib: ModelFormat):
	ii = ImageInput()
	ii.channelOrder = ImageInput.ChannelOrder.RGB
	ii.batchSize = 1

	match dl_lib:
		case ModelFormat.TensorFlow:
			ii.dataOrder = ImageInput.DataOrder.NHWC
		case ModelFormat.PyTorch:
			ii.dataOrder = ImageInput.DataOrder.NCHW



import logging
import os
from pathlib import Path
from typing import override

from ai_parade._toolkit.logging import subprocess_run_log_async

from ._PackageManagerInstaller import PackageManager

logger = logging.getLogger(__name__)


class UV(PackageManager):
	def __init__(self, repo_path: Path):
		self.repo_path = repo_path.expanduser().resolve()

		self._uv_in_repo = [
			Path(os.getcwd()) / ".venv" / "bin" / "uv",
		]
		self._common_options = [
			"--no-progress",
			"--directory",
			self.repo_path,
		]
		self._env = {
			**os.environ.copy(),
			"VIRTUAL_ENV": str(self.repo_path / ".venv"),
			"UV_GIT_LFS": "1",  # enable git lfs
		}

	@override
	def prepend_commands(self, prefix: list[str]):
		self._uv_in_repo = prefix + self._uv_in_repo

	@override
	async def freeze_command(self):
		retCode, stdout, stderr = await subprocess_run_log_async(
			self._uv_in_repo + ["pip", "freeze"] + self._common_options,
			env=self._env,
			logger=logger,
		)
		return stdout

	@override
	async def _create_venv_command(self, python_version: str):
		await subprocess_run_log_async(
			self._uv_in_repo
			+ [
				"venv",
				"--python",
				python_version,
			]
			+ self._common_options,
			env=self._env,
			logger=logger,
		)

	@override
	async def _install_dependencies_command(self, dependencies: list[str]):
		await subprocess_run_log_async(
			self._uv_in_repo
			+ [
				"pip",
				"install",
			]
			+ self._common_options
			+ dependencies,
			env=self._env,
			logger=logger,
		)

	@override
	async def _install_requirements_txt_command(self):
		await subprocess_run_log_async(
			self._uv_in_repo
			+ [
				"pip",
				"install",
				"-r",
				self.repo_path / "requirements.txt",
			]
			+ self._common_options,
			env=self._env,
			logger=logger,
		)

	@override
	async def _install_self_command(self):
		await subprocess_run_log_async(
			self._uv_in_repo
			+ [
				"pip",
				"install",
				"--no-build-isolation",
				self.repo_path,
			]
			+ self._common_options,
			env=self._env,
			logger=logger,
		)



import logging
import traceback
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple

from ai_parade._toolkit.data.ImageInput import ImageInput
from ai_parade._toolkit.get_weights import get_weights_name
from ai_parade._toolkit.metadata.models.final import ModelMetadata
from ai_parade._toolkit.run.ModelRunner import ModelRunner
from ai_parade._toolkit.run.PyTorch import PyTorchRunner

from .converter import Conversion, ConversionResult, Converter

if TYPE_CHECKING:
	import torch  # type: ignore

logger = logging.getLogger(__name__)


class PyTorchConvertBase(ABC):
	def __init__(self, conversion: Conversion, options: Converter.Options):
		self.conversion = conversion
		self.options = options

	class Params(NamedTuple):
		model: "torch.nn.Module"
		sample_inputs: tuple["torch.Tensor"]
		output_weights_path: Path
		default_results: ConversionResult
		metadata: ModelMetadata

	@abstractmethod
	def _export_call(
		self, params: Params
	) -> ConversionResult | list[ConversionResult] | None:
		pass

	def __call__(
		self, runner: ModelRunner
	) -> ConversionResult | list[ConversionResult]:
		import torch  # type: ignore

		assert isinstance(runner, PyTorchRunner)

		match runner.model_metadata.image_input.dataType:
			case ImageInput.DataType.float32:
				dtype = torch.float32
			case ImageInput.DataType.float16:
				dtype = torch.float16
			case ImageInput.DataType.uint8:
				dtype = torch.uint8

		sample_inputs = torch.rand(runner.model_metadata.image_input.shape, dtype=dtype)
		runner.model.eval()

		def run(
			sample_inputs: "torch.Tensor",
		) -> ConversionResult | list[ConversionResult]:
			# run the model to trigger potential lazy initialization
			runner.model(sample_inputs)

			default_results = ConversionResult(
				format=self.conversion[1],
				image_input=runner.model_metadata.image_input,
				path=self.get_output_weights_path(runner),
				output=runner.model_metadata.output,
			)

			results = self._export_call(
				PyTorchConvertBase.Params(
					runner.model,
					(sample_inputs,),
					self.get_output_weights_path(runner),
					default_results=default_results,
					metadata=runner.model_metadata,
				)
			)

			return results or default_results

		try:
			runner.model.to("cpu")
			return run(sample_inputs.to("cpu"))
		except Exception as e:
			msg = str(e)
			if "cpu" in msg or "gpu" in msg or "device" in msg:
				logger.warning("Conversion on CPU model failed, trying cuda instead.")
				logger.info(f"Failed with error: '{msg}'")
				logger.debug(traceback.format_exc())
				runner.model.to("cuda")
				return run(sample_inputs.to("cuda"))
			else:
				raise e

	def get_output_weights_path(self, runner: ModelRunner) -> Path:
		return runner.output_directory / get_weights_name(self.conversion[1])



from ..enums import (
	AIRuntime,
	HardwareAcceleration,
	ModelFormat,
	Platform,
	Quantization,
)
from ._common import ModelAttributes

runtime = AIRuntime.ExecuTorch
capabilities = {
	Platform.Android: {
		Quantization.none: {
			HardwareAcceleration.XNNPACK,
			HardwareAcceleration.Vulkan,
			HardwareAcceleration.QNN,
			# HardwareAcceleration.Exynos
			# HardwareAcceleration.Mediatek
		},
		Quantization.int8: {
			HardwareAcceleration.XNNPACK,
			HardwareAcceleration.Vulkan,
			HardwareAcceleration.QNN,
			# HardwareAcceleration.Exynos
			# HardwareAcceleration.Mediatek
		},
		Quantization.float16: {
			HardwareAcceleration.XNNPACK,
			HardwareAcceleration.Vulkan,
			HardwareAcceleration.QNN,
			# HardwareAcceleration.Exynos
			# HardwareAcceleration.Mediatek
		},
	},
	Platform.IOS: {
		Quantization.none: {
			HardwareAcceleration.XNNPACK,
			# HardwareAcceleration.Metal
			# HardwareAcceleration.CoreML
		},
		Quantization.int8: {
			HardwareAcceleration.XNNPACK,
			# HardwareAcceleration.Metal
			# HardwareAcceleration.CoreML
		},
		Quantization.float16: {
			HardwareAcceleration.XNNPACK,
			# HardwareAcceleration.Metal
			# HardwareAcceleration.CoreML
		},
	},
}

format = ModelFormat.ExecutorTorch
format_distinguishes = [ModelAttributes.Quantization, ModelAttributes.HWAcceleration]



import logging
import re
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import cast

from ai_parade._toolkit.constants import TRACE_LVL
from ai_parade._toolkit.logging import subprocess_run_log_async

logger = logging.getLogger(__name__)


async def check_missing_dependencies(
	root_path: Path, pin_defined: bool, pin_installed: bool
) -> bool:
	"""Checks if there are missing dependencies

	Args:
		root_path: Path to project
		pin_requirements: True if dependencies should be pinned into requirements.txt

	Returns:
		True if there are missing dependencies
	"""
	locked_dependencies = await convert_lock_to_requirements(root_path)
	if locked_dependencies is None:
		logger.warning(
			"Could not resolve lock file/pinned dependencies, trying to generate them from project files."
		)

	requirements_file = root_path / "requirements.txt"
	if pin_installed:
		_, locked_dependencies, _ = await subprocess_run_log_async(
			["uv", "pip", "freeze"], logger=logger
		)
		requirements_file.write_text(locked_dependencies)
	elif pin_defined:
		locked_dependencies = await pin_dependencies(root_path, requirements_file)
	elif locked_dependencies is None:
		with NamedTemporaryFile() as temp_file:
			locked_dependencies = await pin_dependencies(
				root_path, Path(temp_file.name)
			)

	locked_dependencies = locked_dependencies.split("\n")

	logger.debug("Will freeze current dependencies")
	_, actual_dependencies, _ = await subprocess_run_log_async(
		["uv", "pip", "freeze"], logger=logger
	)
	actual_dependencies = actual_dependencies.split("\n")

	difference = set(actual_dependencies) - set(locked_dependencies)
	for dep in difference:
		found_in_lock = next(
			(
				x
				for x in locked_dependencies
				if dep.split("==")[0] in x and not x.strip().startswith("#")
			),
			None,
		)
		logger.warning(
			f"Installed dependency missmatch: found {dep}; need {found_in_lock}"
		)

	return len(difference) > 0


def is_pinned_requirements_txt(requirements_file: Path) -> bool:
	"""Check if given requirements.txt contains only pinned dependencies"""
	with requirements_file.open() as f:
		lines = f.read().splitlines()
		for line in lines:
			line = re.sub(r"#.*$", "", line)  # remove comments
			if line.strip() == "":
				continue
			# the @ denotes git or other VCS dependencies
			# ?todo:  they are not necessarily pinned with this simple check
			if "==" not in line and "@" not in line:
				return False
	return True


async def convert_lock_to_requirements(root_path: Path):
	"""Converts a lock file into a requirements.txt

	Supports:
	- pipenv - Pipfile.lock
	- poetry - poetry.lock
	- requirements.txt

	Uses micropipenv

	Args:
		root_path: Path to project where lock file is

	Returns:
		Contents of the requirements file or its equivalent as a string
	"""
	# See this issue: https://github.com/pytorch/pytorch/issues/140914
	# to why we are importing here and patching sys.modules
	import sys

	import micropipenv
	import typing_extensions

	sys.modules["pip._vendor.typing_extensions"] = typing_extensions

	NO_DEV = True
	ONLY_DIRECT = True

	resolveAs = [  # The order defines priority
		("Pipfile.lock", "pipenv"),
		("poetry.lock", "poetry"),
		("requirements.txt", "requirements.txt"),
	]

	method = None
	for file, resolve_method in resolveAs:
		path = root_path / file
		if path.exists():
			method = resolve_method
			break

	sections = None
	if method == "pipenv":
		sections = micropipenv.get_requirements_sections(
			only_direct=ONLY_DIRECT, no_dev=NO_DEV
		)
	elif method == "poetry":
		sections = micropipenv._poetry2pipfile_lock(  # type: ignore
			only_direct=ONLY_DIRECT, no_dev=NO_DEV
		)
	elif method == "requirements.txt":
		is_pinned_requirements_txt(root_path / "requirements.txt")
		with open(root_path / "requirements.txt", "r") as f:
			return f.read()

	if sections is not None:
		result = cast(
			str,
			micropipenv.requirements_str(
				sections,
				# no_hashes=no_hashes,
				# no_indexes=no_indexes,
				# no_versions=no_versions,
				only_direct=ONLY_DIRECT,
				no_dev=NO_DEV,
				no_comments=True,
			),
		)
		return result
	return None


async def pin_dependencies(
	root_path: Path, requirements_file: Path, pinned_requirements: str | None = None
) -> str:
	"""Pins dependencies

	Args:
		root_path: Path to project
		requirements_file: Path to requirements.txt where dependencies should be pinned to

	Raises:
		FileNotFoundError: when no dependency file could be found

	Returns:
		Pinned dependencies
	"""
	logger.debug(f"Pinning dependencies to `{requirements_file}`")
	resolveAs = [  # The order defines priority
		"requirements.in",
		"pyproject.toml",
		"requirements.txt",
		"setup.py",
		"setup.cfg",
	]

	found_dependency_files: list[Path] = []
	for file in resolveAs:
		path = root_path / file
		if path.exists():
			found_dependency_files.append(path)

	logger.log(TRACE_LVL, "Trying to pin dependencies")
	returnCode, _, stderr = await subprocess_run_log_async(
		[
			"uv",
			"pip",
			"compile",
			"--no-build-isolation",
			"--output-file",
			requirements_file,
		]
		+ (found_dependency_files if pinned_requirements is None else ["-"]),
		check=False,
		logger=logger,
		stdin=None if pinned_requirements is None else pinned_requirements,
	)
	if returnCode != 0:
		raise FileNotFoundError(
			f"Could not resolve dependencies, tried: {', '.join([str(x) for x in found_dependency_files])}",
			stderr,
		)
	logger.debug("Successfully pinned dependencies")
	with open(requirements_file, "r") as f:
		result = f.read()

		return result



from typing import Annotated, Any, Mapping, Optional

from pydantic import (
	BaseModel,
	ConfigDict,
	StringConstraints,
	field_validator,
	model_validator,
)

# backward compatibility <=3.11: typeddict from here
from typing_extensions import TypedDict

from ai_parade._toolkit.data.ImageInput import ImageInput
from ai_parade._toolkit.data.output import Output
from ai_parade._toolkit.enums import ModelFormat
from ai_parade._toolkit.enums.formats import ExactModelFormat
from ai_parade._toolkit.structural_mapping import MappingDefinition

from ..enums import ModelTasks


class InlineCode(TypedDict):
	code: str


NAME_REGEX = "^[-+ _.a-zA-Z0-9]*$"


class ModelMetadataPyproject(BaseModel):
	"""
	pyproject.toml [tool.ai-parade] options for model registration.
	"""

	model_config = ConfigDict(validate_assignment=False, extra="forbid")

	# Basic information about the model #
	#####################################

	name: Annotated[str, StringConstraints(pattern=NAME_REGEX, min_length=1)]
	"""
	Name of the model

	There are some restrictions:
	- it must be unique in the repository
	"""

	family: str | None = None
	"""
	Used to group models.
	"""

	size: float | None = None
	"""
	Size of the model

	Only the ordering is important, not the actual size. 
	"""

	format: ModelFormat | ExactModelFormat
	"""
	Format of the model, e.g. `PyTorch`
	"""

	task: ModelTasks | tuple[ModelTasks, ...]
	"""
	Task or list of tasks which the model solves.
	"""

	task_properties: Mapping[str, object] = {}
	"""
	Additional task properties/specification
	"""

	# Attribution and sources #
	###########################

	citation: Optional[str | tuple[str, ...]] = None
	"""
	Citation(s) in bib format.
	"""

	citation_link: Optional[str | tuple[str, ...]] = None
	"""
	Link(s) for the cited paper(s)
	"""

	download_link: Optional[str] | dict[str, str] = None
	"""
	URL to download weights. Supported services:

		- direct download
		- Google drive

	If the model is trained on multiple datasets this can be a dictionary where the key is the dataset name and the value is the URL to download the weights.
	"""

	license: Optional[str] = None
	"""
	License of the model code and weights in SPDX format if possible.

	see: https://spdx.org/licenses/
	"""

	# Running the model #
	#####################
	install: "InstallOptionsPyproject | None" = None
	"""
	Defines install options.
	"""

	image_input: ImageInput
	"""
	Remarks:
	- resize is with padding, keeping the aspect ratio
	- batch size is not much used
	- before any normalization the input is always [0, 255] for each channel, even with floating point types
	"""

	output: Output
	"""
	Maps output of the model to a dict with `ModelOutput` [see here](data/output.py). The value is recursively string, list or dict. The structure should match the structure of the model output and the leaf nodes values should match the `ModelOutput` keys. For example:

	```toml
	output = ["class_ImageNet_1k", "confidence"]

	output = {label="class_ImageNet_1k", conf="confidence"}
	```

	Special keys:
	- `"...skip"` or "" - skip item in the output list/tuple
	- `"...repeat"` - repeat the same mapping for all items in the output list/tuple
	- `"...repeat as {key}"` - repeat the same mapping for all items in the output list/tuple and put them under the given key in the resulting dict

	Example with special keys:
	```toml
	output = ["class_ImageNet_1k", "...skip", "confidence"]

	output = {bboxes=[["bbox_xmin_normalized", "bbox_xmax_normalized", "bbox_ymin_normalized", "bbox_ymax_normalized"], "...repeat as detection"]}

	output = {label=["class_ImageNet_1k", "...repeat as classification"], conf=["confidence", "...repeat as classification"]}
	```
	"""

	runner_name: Optional[str] = None
	"""
	Name of the custom runner class. This class have to be defined in a python file (`.py`) with the same name and path as the metadata file (i.e. the toml file documented here). This class have to inherit from the `model.ModelRunner` abstract class. Usually it will inherit from other classes like `PyTorchRunner`.

	Setting it `None` - the default value - the python file (if present) will be searched for class with the same name as the model format if found it will be used otherwise the default runner for given format is used.
	"""

	# Additional information #
	##########################
	dataset: Optional[str] | list[str] = None
	"""
	Dataset used to train the model
	"""

	backbone: Optional[str] = None
	"""
	todo:
	Just information about backbone
	"""

	architecture: Optional[str] = None
	"""
	todo:
	Just information about architecture
	"""

	tags: Optional[tuple[str, ...]] = ()
	"""
	Tags
	"""

	# AI Runtime specific options #
	###########################

	pytorch: Optional["PyTorchOptionsPyproject"] = None
	"""
	pytorch specific settings
	"""
	# New AI Runtime: update here

	@model_validator(mode="after")
	def _dl_lib_specific_options(self):
		model_format = (
			self.format if isinstance(self.format, ModelFormat) else self.format.format
		)
		assert self.runner_name is None or getattr(self, model_format.value), (
			f"Model format `{model_format.value}`, no such key is present. "
			"Metadata has to define options specific to the model format."
		)
		return self

	@field_validator("format", "task", mode="before")
	@classmethod
	def _normalize(cls, v: Any):
		if isinstance(v, str):
			return v.lower().strip()
		else:
			return v

	@model_validator(mode="after")
	def _set_install_default(self):
		if self.format == ModelFormat.PyTorch and self.install is None:
			self.install = InstallOptionsPyproject()
		return self


class InstallOptionsPyproject(BaseModel):
	"""
	Options for automatic installation.

	The order of installation is as follows (assuming no steps are skipped):

	- virtual environment initialization
	        - optionally with conda's `environment.yml`
	- installing `install.conda_dependencies` via conda
	- installing `install.pip_dependencies` via pip
	- installing `requirements.txt` or other files via pip
	        - dependencies from `requirements.txt` (pinned and unpinned)
	        - poetry dependencies from `poetry.lock` and `pyproject.toml`
	        - Pipenv dependencies from `Pipfile.lock`

	        The install procedure uses `micropipenv`.
	- install itself via pip
	"""

	model_config = ConfigDict(extra="forbid")

	pip_dependencies: Optional[tuple[str, ...]] = None
	"""
	Additional pip dependencies (other than defined in `requirements.txt`).
	See the `InstallOptions` docs for installation order and precedence
	"""

	conda_dependencies: Optional[tuple[str, ...]] = None
	"""
	Additional conda dependencies (other than `environment.yml`). 
	See the `InstallOptions` docs for installation order and precedence
	"""


class PyTorchOptionsPyproject(BaseModel):
	"""
	Options specific to PyTorch, mainly how to run the model.
	"""

	model_config = ConfigDict(protected_namespaces=(), extra="forbid")

	model_import: Optional[str] = None
	"""
	Name of a module to import. Used with:
	- `model_init_expression`
	- `model_init_fx`
	"""

	model_init_expression: Optional[str] = None
	"""
	Define an expression to run which returns model object (`torch.nn.Module`).
	Returned model should be without initialized weights. The weights will be initialized automatically later.
	Could use these variables:

	- `args: list[cfg]` list of parameters - there is the loaded configuration, if available
	- `module: ModuleType` imported module as defined in `pytorch.model_import`	
	"""

	model_init_fx: Optional[str] = None
	"""
	Name of a method (callable) that will be called to instantiate the model.
	It is given the config (if available) as a positional parameter.
	Returned model should be without initialized weights. The weights will be initialized automatically later.

	Setting `pytorch.model_init_fx = my_function` is equivalent to specifying `pytorch.model_init_expression = "module.my_function(*args)"`.
	"""

	yacs_configs: Optional[tuple[str | InlineCode, ...]] = None
	"""
	List of config files or dict with single key `code`.
	The first entry must be the base configuration python file.
	The other entries are added in the same order as defined. The code has accessible the `cfg` object with the configuration.

	For example:

	```toml
	pytorch.yacs_configs= [
		"base_config.py",
		{code = "cfg = patch_config(cfg)"}
	]
	```
	"""

	load_mapping: Optional[MappingDefinition] = None
	"""
	Maps loaded weights to something.
	It has similar semantic as in `output`. I.E. match the structure of the data in the file and put string `state_dict` on the position of the `state_dict` data.
	"""

	@model_validator(mode="after")
	def _field_set_validation(self):
		assert any(
			[hasattr(self, key) for key in ["model_init_fx", "model_init_expression"]]
		), (
			"No way of constructing model, none of these keys have been defined:"
			"\n\t- `pytorch.model_init_fx`"
			"\n\t- `pytorch.model_init_expression`"
		)
		return self



import logging
from pathlib import Path
from typing import Any, override

from .ModelRunnerBase import (
	Dataset,
	ModelMetadata,
	ModelOutput,
	ModelRunnerBase,
)

logger = logging.getLogger(__name__)


class NotARunner(ModelRunnerBase["NotARunner", "None"]):
	def __init__(self, model_metadata: ModelMetadata, weights: Path) -> None:
		super().__init__(model_metadata, weights)

	@override
	async def __aenter__(self):
		return self

	@override
	async def __aexit__(self, *err: Any):
		self._model = None

	@override
	async def inference(self, dataset: Dataset) -> list[ModelOutput]:
		raise NotImplementedError(
			"Not a runner, this model cannot be used for inference on this device"
		)

	@property
	@override
	def model(self) -> None:
		raise NotImplementedError("Not a runner, there is no model.")



import ai_parade._compat  # noqa: F401
from ai_parade._toolkit.datasets.Dataset import DatasetSplit

from ._toolkit.datasets.loadImageNet import ImageNet, ImageNet_VID
from ._toolkit.datasets.Random import Random

__all__ = ["ImageNet_VID", "Random", "ImageNet", "DatasetSplit"]



from ._toolkit.constants import TRACE_LVL
from ._toolkit.logging import (
	get_ai_parade_logger,
	log_subprocess_result,
	log_subprocess_result_async,
	subprocess_run_log,
	subprocess_run_log_async,
)

__all__ = [
	"TRACE_LVL",
	"log_subprocess_result",
	"log_subprocess_result_async",
	"subprocess_run_log",
	"subprocess_run_log_async",
	"get_ai_parade_logger",
]



import sys

# Backport `override` decorator for Python < 3.12
import typing

if not hasattr(typing, "override"):

	def override(method: typing.Any):
		return method

	typing.override = override

# Backport `StrEnum` for Python < 3.11
import enum

if not hasattr(enum, "StrEnum"):
	from backports.strenum import StrEnum

	enum.StrEnum = StrEnum

# Backport `tomllib` for Python < 3.11
try:
	import tomllib
except ImportError:
	import tomli as tomllib

	# Register the module globally
	sys.modules["tomllib"] = tomllib



import logging
import re
from abc import ABC, abstractmethod
from pathlib import Path
from typing import override

from ai_parade._toolkit.constants import TRACE_LVL, VENV_DIR
from ai_parade._toolkit.metadata.models.final import InstallOptions

from .ModelInstaller import ModelInstaller
from .utilities import convert_lock_to_requirements, is_pinned_requirements_txt

logger = logging.getLogger(__name__)


class PackageManager(ABC):
	"""Abstract base class for package installers (uv, pip, etc.)"""

	@abstractmethod
	def prepend_commands(self, prefix: list[str]):
		"""Prepends all commands with the given prefix"""
		pass

	@abstractmethod
	async def freeze_command(self) -> str:
		"""Returns installed packages in the current environment

		Returns:
			str: installed packages in pip freeze format
		"""
		pass

	@abstractmethod
	async def _create_venv_command(self, python_version: str):
		"""Creates a virtual environment

		Args:
			python_version: Python version to use for the virtual environment
		"""
		pass

	@abstractmethod
	async def _install_dependencies_command(self, dependencies: list[str]):
		"""Installs a list of dependencies"""
		pass

	@abstractmethod
	async def _install_requirements_txt_command(self):
		"""Installs dependencies from requirements.txt"""
		pass

	@abstractmethod
	async def _install_self_command(self):
		"""
		Use `--no-build-isolation` if possible.
		This option in pip does two things:
		- the outside environment is not isolated from the build.
		- the build dependencies are not installed.

		There is an common issue when the package needs to build itself
		with the same version of a dependency as is then using in runtime.
		(i.e. prebuilt torch operations vs torch runtime)
		see: https://github.com/astral-sh/uv/issues/2252#issuecomment-2710054979

		So to make everything hopefully work with the `--no-build-isolation`
		- we are working in a venv (per project)
		- we installed build dependencies from pyproject.toml in the venv prior to calling this
		"""
		pass


class PackageManagerInstaller(ModelInstaller):
	"""Installs dependencies using a package manager

	The installation process is as follows:
	- create venv
	- install pip dependencies:
		- install dependencies from ai-parade.install.pip_dependencies in pyproject.toml
		- install dependencies from requirements.txt
		- install self, this will invoke build scripts like setup.py, pyproject.toml, etc.
	"""

	def __init__(
		self,
		repo_path: Path,
		install_options: InstallOptions,
		package_manager: PackageManager,
	):
		"""Interacts with a model

		Args:
			repo_path: Path to the repository
			package_manager: The package manager to use (uv, pip, etc.)
			install_options: Optional installation options. If None, will create a new one with default values.
		"""
		self.repo_path = repo_path
		self.package_manager = package_manager
		self.venv_path = repo_path / VENV_DIR
		self.install_parameters = install_options

	def is_venv_created(self) -> bool:
		"""Checks if venv could be found

		Returns:
			True if venv is found
		"""
		logger.debug("Checking if venv is created")
		has_venv = (self.venv_path / "bin" / "activate").is_file()
		logger.log(TRACE_LVL, f"venv is {'found' if has_venv else 'not found'}")
		return has_venv

	async def ensure_requirements_txt(self):
		requirements_txt = self.repo_path / "requirements.txt"
		if not requirements_txt.is_file():
			result = await convert_lock_to_requirements(requirements_txt.parent)
			if result is None:
				raise FileNotFoundError(
					"requirements.txt could not be found or created."
				)
			else:
				with open(Path(requirements_txt.name), "w") as f:
					f.write(result)
				return True
		return True

	async def is_pip_dependencies_installed(self) -> tuple[bool, bool, bool, bool]:
		"""Checks if pip dependencies are installed

		Returns tuple of bools where each one represents one check:
		- dependencies from requirements.txt
		- self
		- dependencies from ai-parade.install.pip_dependencies in pyproject.toml
		- dependencies from build-system.requires in pyproject.toml

		Returns:
			tuple of bools where each one represents one check

			`(bool, bool, bool, bool)`: (requirements.txt, self, pip_dependencies, build_system_dependencies)
		"""
		logger.debug("Checking if pip dependencies are installed")
		freezed = await self.package_manager.freeze_command()

		requirements_txt = self.repo_path / "requirements.txt"

		installed_requirements = False
		if await self.ensure_requirements_txt() and requirements_txt.is_file():
			with requirements_txt.open() as f:
				installed_requirements = True
				for dependency in f.read().splitlines():
					dependency = dependency.strip()
					if dependency.startswith("#"):
						continue
					dependency = dependency.split("#")[0]
					if dependency not in freezed:
						installed_requirements = False
						logger.debug(f"Dependency {dependency} not installed")
		logger.log(TRACE_LVL, f"installed_requirements: {installed_requirements}")

		installed_self = freezed.find("@ file://") != -1
		logger.log(TRACE_LVL, f"installed_self: {installed_self}")

		def is_in_freeze(dependencies: list[str] | tuple[str, ...]) -> bool:
			return all(
				[
					re.split("@|<|<=|!=|==|>=|>|~=|===", dependency)[0].strip()
					in freezed
					for dependency in dependencies
				]
			)

		if self.install_parameters.pip_dependencies is None:
			installed_list = True
		else:
			installed_list = is_in_freeze(self.install_parameters.pip_dependencies)
		logger.log(TRACE_LVL, f"installed_list: {installed_list}")

		installed_build_dependencies = is_in_freeze(
			self.install_parameters.build_system_dependencies
		)
		logger.log(
			TRACE_LVL, f"installed_build_dependencies: {installed_build_dependencies}"
		)

		return (
			installed_requirements,
			installed_self,
			installed_list,
			installed_build_dependencies,
		)

	@override
	async def is_installed(self) -> bool:
		return (
			self.repo_path.exists()
			and self.is_venv_created()
			and all(await self.is_pip_dependencies_installed())
		)

	async def _create_venv(self):
		"""Creates python venv"""
		if self.is_venv_created():
			logger.debug("Virtual environment already there. Skipping")
			return

		await self.package_manager._create_venv_command(self.install_parameters.python)

	async def _install_pip_dependencies(self):
		"""Installs pip dependencies

		These include:
		- dependencies from ai-parade.install.pip_dependencies in pyproject.toml
		- dependencies from requirements.txt
		- build dependencies from pyproject.toml (`build-system.requires`)
		- self
		"""
		(
			installed_requirements,
			installed_self,
			installed_list,
			installed_build_dependencies,
		) = await self.is_pip_dependencies_installed()

		if self.install_parameters.pip_dependencies is not None:
			if installed_list:
				logger.debug("Custom pip dependencies already installed. Skipping")
			else:
				logger.debug("Installing pip dependencies")
				await self.package_manager._install_dependencies_command(
					list(self.install_parameters.pip_dependencies)
				)

		if installed_requirements:
			logger.debug(
				"Pip dependencies from requirements.txt already installed. Skipping"
			)
		else:
			if not (self.repo_path / "requirements.txt").exists():
				raise FileNotFoundError("No requirements.txt found in the repository")
			if not is_pinned_requirements_txt(self.repo_path / "requirements.txt"):
				raise ValueError("requirements.txt is not pinned")

			logger.debug("Installing pip dependencies from requirements.txt")
			await self.package_manager._install_requirements_txt_command()

		if len(self.install_parameters.build_system_dependencies) != 0:
			if installed_build_dependencies:
				logger.debug("Build dependencies already installed. Skipping")
			else:
				logger.debug("Installing build dependencies")
				await self.package_manager._install_dependencies_command(
					self.install_parameters.build_system_dependencies
				)

		if installed_self:
			logger.debug("The project is already installed. Skipping")
		else:
			if (self.repo_path / "setup.py").exists() or (
				self.repo_path / "pyproject.toml"
			).exists():
				logger.debug("Running pip install on self")
				await self.package_manager._install_self_command()

	@override
	async def install(self, override: bool = False):
		"""Installs model

		Args:
			override: If True, will override all existing files. Defaults to False.
		"""
		logger.info("Installing model")

		if override:
			# todo: nuke the dependencies on override == True
			raise NotImplementedError
		logger.debug("Creating virtual environment.")
		await self._create_venv()

		logger.debug("Installing pip dependencies.")
		await self._install_pip_dependencies()
		logger.info("Installation done.")



from inspect import getmembers, isclass
from itertools import chain
from typing import Any, Callable, TypeVar

import numpy as np

from ai_parade._toolkit.data.output import ModelOutput

TType = TypeVar("TType")


def reduce_list(
	my_list: list,
	reduce_function: Callable[[TType, TType], TType],
	list_as_tuples: bool = False,
):
	my_list = my_list.copy()
	if len(my_list) < 2:
		return my_list
	i = 0
	for i in range(1, len(my_list)):
		my_list[i] = reduce_objects(
			my_list[i], my_list[i - 1], reduce_function, list_as_tuples
		)
	return my_list[i]


def reduce_objects(
	object1: Any,
	object2: Any,
	reduce_function: Callable[[TType, TType], TType],
	list_as_tuples: bool = True,
):
	if (
		getattr(object1, "__irreducible__", False)
		or getattr(object2, "__irreducible__", False)
	) or (
		isinstance(object1, (int, float, np.number))
		and isinstance(object2, (int, float, np.number))
	):
		return reduce_function(object1, object2)  # type: ignore
	if isinstance(object1, (list, tuple, np.ndarray)) and isinstance(
		object2, (list, tuple, np.ndarray)
	):
		if list_as_tuples:
			assert len(object1) == len(object2)
			return [
				reduce_objects(value1, value2, reduce_function, list_as_tuples)
				for value1, value2 in zip(object1, object2)
			]
		else:
			return [
				reduce_list(
					list(chain(object1, object2)), reduce_function, list_as_tuples
				)
			]
	elif isinstance(object1, dict) and isinstance(object2, dict):
		assert object1.keys() == object2.keys()
		return {
			key: reduce_objects(
				object1[key], object2[key], reduce_function, list_as_tuples
			)
			for key in object1.keys()
		}
	elif (type(object1) is type(object2)) and (isclass(object1) and isclass(object2)):
		return {
			key: reduce_objects(
				value1, getattr(object2, key), reduce_function, list_as_tuples
			)
			for key, value1 in getmembers(
				object1,
				lambda x: not callable(x[1]) and not x[0].startswith("__"),
			)
		}
	else:
		raise ValueError(f"Unsupported type: {type(object1)}")


def compute_statistics(
	output_list1: list[ModelOutput], output_list2: list[ModelOutput]
):
	diff_list = [
		reduce_objects(output1, output2, lambda x, y: abs(x - y))
		for output1, output2 in zip(output_list1, output_list2)
	]

	max_diff = reduce_list(diff_list, max)
	min_diff = reduce_list(diff_list, min)

	def sum_count(x: int | float | list[float], y: int | float | list[float]):
		if not isinstance(x, list):
			x = [x, 1]
		if not isinstance(y, list):
			y = [y, 1]
		return IrreducibleList([x[0] + y[0], x[1] + y[1]])

	sum_diff = reduce_list(diff_list, sum_count)
	mean_diff = reduce_objects(sum_diff, sum_diff, lambda x, y: x[0] / x[1])

	return {
		"max_difference": max_diff,
		"min_difference": min_diff,
		"mean_difference": mean_diff,
	}


def irreducible(cls: type):
	cls.__irreducible__ = True
	return cls


@irreducible
class IrreducibleList(list):
	pass



import logging
import platform
from pathlib import Path
from typing import Any, cast

import tomllib
from pydantic import ValidationError

from ai_parade._toolkit.constants import TRACE_LVL
from ai_parade._toolkit.enums.formats import ExactModelFormat
from ai_parade._toolkit.install._installer_selector import choose_installer

from .custom import Customizations, get_customizations
from .models.api import ModelMetadataApi
from .models.final import (
	InstallOptions,
	ModelMetadata,
)
from .models.pyproject import ModelMetadataPyproject

logger = logging.getLogger(__name__)

ModelMetadataApi.model_rebuild()


def _deep_dict_merge(base: dict, variant: dict):
	"""Recursively merges two dictionaries, overwriting the base with the variant"""

	def merge_recursion(base: Any, variant: Any):
		if isinstance(base, dict) and isinstance(variant, dict):
			for key in variant.keys():
				if key not in base or not merge_recursion(base[key], variant[key]):
					base[key] = variant[key]
			return True
		elif (
			isinstance(base, list)
			and isinstance(variant, list)
			and len(base) == len(variant)
		):
			for key, _ in enumerate(base):
				if not merge_recursion(base[key], variant[key]):
					base[key] = variant[key]
				return True
		else:
			return False

	base = base.copy()
	merge_recursion(base, variant)
	return base


_VARIANTS_KEY = "variants"


def _parse_variants(raw: dict):
	"""Parses the variants from the raw metadata

	Resolves overriding of the base (non-variant) properties, assign unique names to the variants
	"""
	parsed: list[ModelMetadataPyproject] = []
	common = raw.copy()
	del common[_VARIANTS_KEY]

	logger.info(f"Found metadata with {len(raw['variants'])} variants")
	for i, variant in enumerate(raw["variants"]):
		parsed_variant = _deep_dict_merge(common, variant)
		definition = ModelMetadataPyproject.model_validate(parsed_variant)
		logger.debug(f"Found variant `{definition.name}`")
		logger.log(TRACE_LVL, f"--> {definition}")
		parsed.append(definition)
	return parsed


class MetadataNotFoundError(Exception):
	"""
	Exception raised when no model metadata has been found
	"""

	pass


def _raw_to_pyproject(
	raw: dict, group_name: str, has_customizations: bool
) -> list[ModelMetadataPyproject]:
	"""Parses the raw metadata from the pyproject.toml file into ModelMetadataDefinitions"""
	parsed = []
	if "tool" not in raw or "ai-parade" not in raw["tool"]:
		if not has_customizations:
			raise MetadataNotFoundError(group_name)
	else:
		raw_metadata = raw["tool"]["ai-parade"]

		try:
			if _VARIANTS_KEY in raw_metadata:
				parsed = _parse_variants(raw_metadata)
			else:
				parsed = [ModelMetadataPyproject.model_validate(raw_metadata)]
		except ValidationError as e:
			e.add_note(group_name)
			raise e
	return parsed


def _pyproject_to_api(
	definitions: list[ModelMetadataPyproject],
) -> list[ModelMetadataApi]:
	"""Converts ModelMetadataApi to ModelMetadataDefinitions"""
	return [
		ModelMetadataApi(
			**{
				**definition.model_dump(),
				"format": (
					definition.format
					if isinstance(definition.format, ExactModelFormat)
					else ExactModelFormat(definition.format)
				),
			},
		)
		for definition in definitions
	]


def _customization_to_api(customizations: Customizations) -> list[ModelMetadataApi]:
	"""Loads ModelMetadataApi from Customizations"""
	customization_metadata = []
	if customizations.include is None:
		return []
	else:
		logger.info("Loading metadata from Customizations")
		for api in customizations.include():
			assert isinstance(api, ModelMetadataApi)
			customization_metadata.append(api)
			logger.debug(f"Customizations added model: {api.name}")
			logger.log(TRACE_LVL, f"--> {api}")
		return customization_metadata


def _api_to_final(
	api: ModelMetadataApi,
	repo_path: Path | None,
	customizations: Customizations,
	build_dependencies: list[str],
	python_version: str,
) -> ModelMetadata:
	"""
	Converts ModelMetadataApi object to ModelMetadata by:
	- setting default values on unset properties
	- setting generated properties
	"""
	api.get_installer = api.get_installer or choose_installer(
		api.install is not None,
		customizations,
		False if api.install is None else api.install.conda_dependencies is not None,
	)
	parsed_definition_dict = api.model_dump()

	if api.install is not None:
		parsed_definition_dict["install"] = InstallOptions(
			**api.install.model_dump(),
			build_system_dependencies=build_dependencies,
			python=python_version,
		).model_dump()  # ?todo: why it does not work without it???

	return ModelMetadata(
		**parsed_definition_dict,
		repository_path=repo_path,
	)


def parse_metadata(
	raw: dict | None,
	*,
	repo_path: Path | None,
	customizations: Customizations,
	# family_name: str | None,
	python_version: str | None = None,
) -> list[ModelMetadata]:
	"""
	Parses the raw metadata dict into ModelMetadata.

	Args:
		raw: raw metadata dict
		repo_path: path to the repository
		customizations: Customizations

	Returns:
		list[ModelMetadata]: list of ModelMetadata
	"""
	raw = raw or {}
	debug_name = repo_path.stem if repo_path else "unknown"
	python_version = python_version or platform.python_version()

	metadataDefinitions = _raw_to_pyproject(
		raw, debug_name, customizations.include is not None
	)
	metadataApis = _pyproject_to_api(metadataDefinitions)
	metadataApis.extend(_customization_to_api(customizations))

	# Set generated field's defaults
	parsed_list: list[ModelMetadata] = []
	for api in metadataApis:
		build_dependencies = []
		if (
			api.install is not None
			and "build-system" in raw
			and "requires" in raw["build-system"]
		):
			build_dependencies = cast(list[str], raw["build-system"]["requires"])
		else:
			# if no build-system is defined, add setuptools
			build_dependencies.append("setuptools")
		final_metadata = _api_to_final(
			api, repo_path, customizations, build_dependencies, python_version
		)

		parsed_list.append(final_metadata)
	return parsed_list


async def get_local_metadata(
	repo_path: Path,
) -> list[ModelMetadata]:
	"""
	Get metadata from the pyproject.toml file in the repository

	Args:
		repo_path: Path to the repository

	Returns:
		list[ModelMetadata]: List of found ModelMetadata
	"""
	logger.info(f"Getting metadata from `{repo_path}`")
	pyproject = repo_path / "pyproject.toml"

	if pyproject.exists():
		logger.info(f"Parsing {pyproject}")
		with pyproject.open("rb") as file:
			raw = tomllib.load(file)
	else:
		logger.info(f"pyproject.toml not found in {repo_path}, will use customizations")
		raw = {}

	python_version = raw.get("required-python", platform.python_version())
	install_options = InstallOptions(
		pip_dependencies=tuple(raw.get("dependency-groups", {}).get("ai-parade", [])),
		python=python_version,
		build_system_dependencies=list(
			raw.get("build-system", {}).get("requires", ["setuptools"])
		),
	)
	customizations = await get_customizations(repo_path, install_options)

	return parse_metadata(
		raw,
		repo_path=repo_path,
		customizations=customizations,
		python_version=python_version,
	)



import sys
import torch
import torch.nn as nn


# Define a simple CNN using Sequential
class SimpleCNN(nn.Module):
	def __init__(self):
		super(SimpleCNN, self).__init__()
		self.model = nn.Sequential(
			nn.Flatten(),
			nn.Linear(3 * 4 * 4, 1),
		)

	def forward(self, x):
		assert sys.version_info.major == 3
		assert sys.version_info.minor == 10

		return self.model(x)


def mock():
	model = SimpleCNN()
	for m in model.modules():
		if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
			nn.init.uniform_(m.weight, a=-1.0, b=1.0)
			nn.init.uniform_(m.bias, a=-1.0, b=1.0)

	print("running")
	model.eval()
	model(torch.ones(1, 3, 4, 4))

	print("saving")
	torch.save(model.state_dict(), "weights.pth")


if __name__ == "__main__":
	mock()



from pathlib import Path

import pytest
from typer.testing import CliRunner

from ai_parade._cli.ai_parade.main import app
from ai_parade._toolkit.converters.converter import ConversionResult

from ._common import files_equal, setup_cwd_any

runner = CliRunner()


class TestConvertCompare:
	repo_path = Path("fixtures/FooBarModel")

	@pytest.fixture(autouse=True)
	def copy_repo(self, tmp_path, monkeypatch):
		setup_cwd_any(
			tmp_path,
			monkeypatch,
			repo_path=self.repo_path,
			skip=["models"],
		)

	@pytest.mark.parametrize(
		["args", "file_names"],
		[
			pytest.param(*p, id=p[1])
			for p in [
				# LiteRT
				(
					"convert weights.pth FooBarModel tflite --artifacts build --separate-env",
					"tflite.float32.all.tflite",
				),
				(
					"convert weights.pth FooBarModel tflite --quantization float16 --separate-env",
					"tflite.float16.all.tflite",
				),
				(
					"convert weights.pth FooBarModel tflite --quantization int8 --calibration-dataset Random --separate-env",
					"tflite.int8.all.tflite",
				),
				# ncnn
				(
					"convert weights.pth FooBarModel ncnn --separate-env",
					"ncnn.float32.all.ncnn",
				),
				(
					"convert weights.pth FooBarModel ncnn --quantization float16 --separate-env",
					"ncnn.float16.all.ncnn",
				),
				(
					"convert weights.pth FooBarModel ncnn --quantization int8 --calibration-dataset Random --separate-env",
					"ncnn.int8.all.ncnn",
				),
				# Executorch
				(
					"convert weights.pth FooBarModel Executorch --hw-acceleration XNNPACK --separate-env",
					"executorch.float32.XNNPACK.pte",
				),
				(
					"convert weights.pth FooBarModel Executorch --quantization float16 --hw-acceleration XNNPACK --separate-env",
					"executorch.float16.XNNPACK.pte",
				),
				(
					"convert weights.pth FooBarModel Executorch --quantization int8 --calibration-dataset Random --hw-acceleration XNNPACK --separate-env",
					"executorch.int8.XNNPACK.pte",
				),
				(
					"convert weights.pth FooBarModel Executorch --hw-acceleration Vulkan --separate-env",
					"executorch.float32.Vulkan.pte",
				),
				(
					"convert weights.pth FooBarModel Executorch --quantization float16 --hw-acceleration Vulkan --separate-env",
					"executorch.float16.Vulkan.pte",
				),
				(
					"convert weights.pth FooBarModel Executorch --quantization int8 --calibration-dataset Random --hw-acceleration Vulkan --separate-env",
					"executorch.int8.Vulkan.pte",
				),
				(
					"convert weights.pth FooBarModel Executorch --hw-acceleration QNN --separate-env",
					"executorch.float32.QNN.pte",
				),
				(
					"convert weights.pth FooBarModel Executorch --quantization float16 --hw-acceleration QNN --separate-env",
					"executorch.float16.QNN.pte",
				),
				(
					"convert weights.pth FooBarModel Executorch --quantization int8 --calibration-dataset Random --hw-acceleration QNN --separate-env",
					"executorch.int8.QNN.pte",
				),
				# ONNX
				(
					"convert weights.pth FooBarModel ONNX --separate-env",
					"onnx.float32.all.onnx",
				),
				(
					"convert weights.pth FooBarModel ONNX --quantization float16 --separate-env",
					"onnx.float16.all.onnx",
				),
				(
					"convert weights.pth FooBarModel ONNX --quantization int8 --calibration-dataset Random --separate-env",
					"onnx.int8.all.onnx",
				),
			]
		],
	)
	def test_conversion(self, args, file_names):
		result = runner.invoke(app, args)

		assert result.exit_code == 0

		for file_name in file_names if isinstance(file_names, list) else [file_names]:
			# check model
			expected_model: Path = (
				Path(__file__).parent.parent / self.repo_path / "models" / file_name
			)
			actual_model = Path(file_name)

			files_equal(expected_model, actual_model)

			# check json conversion results
			expected_result = expected_model.with_name(expected_model.name + ".json")
			actual_result = actual_model.with_name(actual_model.name + ".json")

			actual_result = ConversionResult.model_validate_json(
				actual_result.read_text()
			)
			expected_result = ConversionResult.model_validate_json(
				expected_result.read_text()
			)

			assert actual_result == expected_result



from ai_parade._toolkit.metadata.custom import Customizations
from ai_parade._toolkit.metadata.models.api import InstallerGetter

from ._PackageManagerInstaller import PackageManagerInstaller
from .ModelInstaller import NOPInstaller


def choose_installer(
	install_options: bool = False,
	customizations: Customizations | None = None,
	has_conda: bool = False,
) -> InstallerGetter:
	"""Choose appropriate installer based on parameters

	Args:
		install_options: True if tools.ai-parade.install is present. Defaults to False.
		customizations: Customizations. Defaults to None.
		has_conda: True if tools.ai-parade.c.install.conda_dependencies is present. Defaults to False.

	Returns:
		Selected installer in form of InstallerGetter
	"""

	if customizations is not None and len(customizations.installers) > 0:
		installer = next(iter(customizations.installers.values()))
		return installer
	else:
		if not install_options:
			return NOPInstaller
		else:
			if has_conda:
				from ._CondaInstaller import CondaInstaller
				from ._Uv import UV

				def conda_installer(metadata):
					assert metadata.repository_path is not None
					return CondaInstaller(
						metadata,
						PackageManagerInstaller(
							metadata.repository_path,
							metadata.install,
							UV(metadata.repository_path),
						),
					)

				return conda_installer
			else:
				from ._Uv import UV

				def uv_installer(metadata):
					assert metadata.repository_path is not None
					return PackageManagerInstaller(
						metadata.repository_path,
						metadata.install,
						UV(metadata.repository_path),
					)

				return uv_installer



from pathlib import Path

from typer.testing import CliRunner

from ai_parade._cli.ai_parade.main import app

from ._common import setup_cwd_any

runner = CliRunner()


def test_python_3_10_separate_env(tmp_path, monkeypatch):
	setup_cwd_any(
		tmp_path,
		monkeypatch,
		repo_path=Path("fixtures/EnvCheckerModel"),
	)
	result = runner.invoke(app, "inference weights.pth EnvCheckerModel --separate-env")
	assert result.exit_code == 0


def test_python_3_10_no_separate_env(tmp_path, monkeypatch):
	setup_cwd_any(
		tmp_path,
		monkeypatch,
		repo_path=Path("fixtures/EnvCheckerModel"),
	)
	result = runner.invoke(app, "inference weights.pth EnvCheckerModel")
	assert result.exit_code != 0



import logging
from inspect import isclass
from pathlib import Path
from typing import Callable

from ai_parade._toolkit.constants import CUSTOM_DEFINITIONS_FILE_NAME, TRACE_LVL
from ai_parade._toolkit.install._PackageManagerInstaller import PackageManagerInstaller
from ai_parade._toolkit.install._Uv import UV
from ai_parade._toolkit.install.ModelInstaller import ModelInstaller
from ai_parade._toolkit.metadata.models.api import (
	InstallerGetter,
	ModelMetadataApi,
	RunnerGetter,
)
from ai_parade._toolkit.metadata.models.final import InstallOptions
from ai_parade._toolkit.run.ModelRunner import ModelRunner
from ai_parade._toolkit.sys_path_checkpoint import SysPathCheckpoint

logger = logging.getLogger(__name__)


class Customizations:
	"""
	Customizations object

	`ai-parade.py` file specification:
	- include: () -> list[ModelMetadataApi]
		Function to define metadata instead of using `pyproject.toml`.
	- installers: (ModelMetadata) -> ModelInstaller
		Model installer class.
		Name of the class could be used to select the installer from within the model metadata.
		The parameters are passed to the constructor.
	- runners: (ModelMetadata) -> ModelRunner
		Model runner class.
		Name of the class could be used to select the runner from within the model metadata.
		The parameters are passed to the constructor.
	"""

	include: Callable[[], list[ModelMetadataApi]] | None = None
	installers: dict[str, InstallerGetter] = {}
	runners: dict[str, RunnerGetter] = {}


async def get_customizations(
	repo_path: Path,
	install_options: InstallOptions,
):
	"""Get customizations from the repository, **installs the model dependencies**.

	The customizations are defined in the `ai-parade.py` file in the repository.

	Args:
		repo_path: Path to the repository

	Returns:
		Customizations: Customizations object with the customizations
	"""
	# todo: this loads everything (modules) into the memory, which is bad
	# best would be to run everything in a separate process via puppet, and implement sw for that
	customizations = Customizations()
	customizations_file = repo_path / CUSTOM_DEFINITIONS_FILE_NAME
	if customizations_file.exists():
		import importlib.util
		import sys

		logger.info(f"Customization file found: {customizations_file}")

		installer = PackageManagerInstaller(repo_path, install_options, UV(repo_path))
		if not await installer.is_installed():
			logger.info(
				"Will install model dependencies, to be able to run customizations"
			)
			await installer.install()

		logger.debug("Importing customizations (ai-parade.py)")

		with SysPathCheckpoint(repo_path):
			# patch path so it includes customizations and its dependencies
			sys.path.insert(0, str(repo_path / ".venv/lib/python3.10/site-packages"))
			sys.path.insert(0, str(repo_path))

			spec = importlib.util.spec_from_file_location(
				"ai_parade_customizations", customizations_file
			)
			custom_definitions_module = importlib.util.module_from_spec(spec)  # type: ignore
			spec.loader.exec_module(custom_definitions_module)  # type: ignore

		logger.debug("Inspecting customizations (ai-parade.py):")

		# includes
		customizations.include = getattr(custom_definitions_module, "include", None)
		if customizations.include is not None:
			assert callable(customizations.include)
			logger.debug("Found include function")

		for x in dir(custom_definitions_module):
			member = getattr(custom_definitions_module, x)
			if (
				not hasattr(member, "__module__")
				or not member.__module__ == custom_definitions_module.__name__
			):
				logger.log(TRACE_LVL, f"Skipping {x}")
				continue

			# runners
			if isclass(member) and issubclass(member, ModelRunner):
				customizations.runners[x] = member
				logger.debug(f"Found runner `{x}`")

			# installers
			if isclass(member) and issubclass(member, ModelInstaller):
				customizations.installers[x] = member
				logger.debug(f"Found installer `{x}`")
	return customizations



import itertools
import logging
import random
from abc import ABC, abstractmethod
from copy import copy as copy_object
from enum import Enum
from functools import partial
from typing import (
	Any,
	Callable,
	Generic,
	Iterable,
	Iterator,
	NamedTuple,
	Sequence,
	TypeVar,
	override,
)

import numpy as np

from ai_parade._toolkit.constants import TRACE_LVL
from ai_parade._toolkit.data.ImageInput import ImageInput
from ai_parade._toolkit.data.output import ModelOutput
from ai_parade._toolkit.data.preprocessing import preprocess_fx

logger = logging.getLogger(__name__)


class DatasetInfo(NamedTuple):
	source_data_shape: Sequence[int]
	data_shape: Sequence[int]
	data_source: str


Output_batched = tuple[np.ndarray, Sequence[ModelOutput], Sequence[DatasetInfo]]
Output_unbatched = tuple[np.ndarray, ModelOutput, DatasetInfo]


TWalkOutput = TypeVar("TWalkOutput")


class DatasetSplit(Enum):
	train = "train"
	val = "val"
	test = "test"


class Dataset(ABC, Iterable[Output_batched], Generic[TWalkOutput]):
	"""Abstract class for datasets

	This class implements common functionality for datasets, such as
	- preprocessing, map
	- skip, take
	- shuffle
	- adding metadata of its source

	Remarks:
		The dataset must be picklable. This mean we can't store generators or lambdas :-(
		As a workaround we can put them in functions and create them when needed

	"""

	def __init__(self):
		self._iterable_transformations: list[
			Callable[[Iterable[Output_unbatched]], Iterable[Output_unbatched]]
		] = []
		self._preload_transformations: list[
			Callable[[Iterable[Any]], Iterable[Any]]
		] = []

		self._batch_size: int = 1
		self._skip_last_batch = False

	@abstractmethod
	def walk(self) -> Iterable[TWalkOutput]:
		"""Walk through the dataset and yield data

		This should be cheap and stable operation.
		The data loading should be done in the parse_data method.

		Returns:
			Iterable of data
		"""
		pass

	@abstractmethod
	def parse_data(
		self, iterable: Iterable[TWalkOutput], should_load: bool
	) -> Iterable[Output_unbatched]:
		"""Parse the data from the iterable and yield data, annotations and info

		Args:
			iterable: Iterable with format of walk method
			should_load: Whether to load the data, or load just annotations

		Returns:
			Ready to use data
		"""
		pass

	@override
	def __iter__(
		self,
	) -> Iterator[Output_batched]:
		return self._iter_implementation(batching=True)  # type: ignore

	def no_batch_iterator(
		self,
	) -> Iterator[Output_unbatched]:
		"""Return an iterator without batching

		Returns:
			Iterator of unbatched data
		"""
		return self._iter_implementation(batching=False)  # type: ignore

	def no_load_iterator(
		self,
	) -> Iterator[Output_unbatched]:
		"""Return an iterator of only annotations

		Returns:
			Iterator of unbatched annotations without data
		"""
		return self._iter_implementation(batching=False, should_load=False)  # type: ignore

	def _iter_implementation(
		self, batching: bool, should_load: bool = True
	) -> Iterator[Output_batched | Output_unbatched]:
		logger.debug("Created iterator")
		logger.log(
			TRACE_LVL,
			"Iterator with transformations: \n"
			f"{self._preload_transformations}\n"
			f"{self._iterable_transformations}\n",
		)

		iterable = self.walk()

		# run data independent transformations i.e. take, skip
		for transform in self._preload_transformations:
			iterable = transform(iterable)

		# load the data
		iterable = self.parse_data(iterable, should_load)  # type: ignore

		# run data dependent transformations
		for transform in self._iterable_transformations:
			iterable = transform(iterable)

		iterable = self._set_final_info(iterable)

		if batching:
			iterable = self._batched(iterable)

		for x in iterable:
			yield x

	def _set_final_info(
		self, iterable: Iterable[Output_unbatched]
	) -> Iterable[Output_unbatched]:
		for data, annotations, info in iterable:
			info = info._replace(data_shape=data.shape)
			yield data, annotations, info

	def _batched(
		self, iterable: Iterable[Output_unbatched]
	) -> Iterable[Output_batched]:
		assert self._batch_size is not None
		# save current values in the iterator
		batch_size = self._batch_size
		skip_last_batch = self._skip_last_batch

		batch_data = []
		batch_annotations = []
		batch_info = []
		for data, annotations, info in iterable:
			batch_data.append(data)
			batch_annotations.append(annotations)
			batch_info.append(info)
			if len(batch_data) == batch_size:
				yield np.array(batch_data), batch_annotations, batch_info
				batch_data = []
				batch_annotations = []
				batch_info = []

		if len(batch_data) != 0 and not skip_last_batch:
			yield np.array(batch_data), batch_annotations, batch_info

	def show(self, n: int | None = None) -> Output_batched | list[Output_batched]:
		"""Show the first n elements of the dataset"

		Args:
			n: number of elements to show. If None, show only the first element
		Returns:
			First n elements of the dataset or the first element if n is None
		"""
		stop = n or 1
		result = list(itertools.islice(self, stop))
		return result[0] if n is None else result

	def take(self, n: int):
		"""Take the first n elements of the dataset

		Args:
			n: number of elements to take
		Returns:
			Dataset with the first n elements
		"""
		copy = copy_object(self)
		copy._preload_transformations = self._preload_transformations.copy()
		copy._preload_transformations.append(partial(self._islice, start=0, stop=n))
		return copy

	@staticmethod
	def _shuffle(
		iterable: Iterable[Output_unbatched], seed: int | None
	) -> Iterable[Output_unbatched]:
		if seed is not None:
			random.seed(seed)
		shuffled = list(iterable)
		random.shuffle(shuffled)
		return shuffled

	def shuffle(self, seed: int | None = None):
		"""Shuffle the dataset

		Args:
			seed: seed for the random number generator
		Returns:
			Shuffled dataset
		"""
		copy = copy_object(self)
		copy._iterable_transformations = self._iterable_transformations.copy()
		copy._iterable_transformations.append(partial(self._shuffle, seed=seed))
		return copy

	@staticmethod
	def _map(
		iterable: Iterable[Output_unbatched], fx: Callable[[np.ndarray], np.ndarray]
	):
		for data, annotations, info in iterable:
			yield fx(data), annotations, info

	def map(self, fx: Callable[[np.ndarray], np.ndarray]):
		"""Transform the data with a function

		Args:
			fx: function to transform each element of the dataset
		Returns:
			Transformed dataset
		"""
		copy = copy_object(self)
		copy._iterable_transformations = self._iterable_transformations.copy()
		copy._iterable_transformations.append(partial(self._map, fx=fx))
		return copy

	def preprocess(self, image_input: ImageInput):
		"""Convert data to the format of the model

		Args:
			image_input: image input to convert the data to
		Returns:
			Transformed dataset
		"""

		return self.map(preprocess_fx(image_input)).set(
			batch_size=image_input.batchSize
		)

	@staticmethod
	def _islice(iterable: Iterable, start: int | None, stop: int | None):
		return itertools.islice(iterable, start, stop)

	def skip(self, n: int):
		"""Skip the first n elements of the dataset

		Args:
			n: number of elements to skip
		Returns:
			Dataset with the first n elements skipped
		"""
		copy = copy_object(self)
		copy._preload_transformations = self._preload_transformations.copy()
		copy._preload_transformations.append(partial(self._islice, start=n, stop=None))
		return copy

	@property
	def batch_size(self):
		return self._batch_size

	@property
	def skip_last_batch(self):
		return self._skip_last_batch

	def set(
		self, *, batch_size: int | None = None, skip_last_batch: bool | None = None
	) -> "Dataset":
		"""Set dataset properties

		Args:
			batch_size: size of the batch
			skip_last_batch: skip the last batch if it is smaller than the batch size
		Returns:
			Dataset with the batch size and skip_last_batch set
		"""
		copy = copy_object(self)
		copy._batch_size = batch_size or self._batch_size
		copy._skip_last_batch = skip_last_batch or self._skip_last_batch
		return copy



import logging
from pathlib import Path
from typing import override

from ai_parade._toolkit.logging import subprocess_run_log_async

from ._PackageManagerInstaller import PackageManager

logger = logging.getLogger(__name__)


class Pip(PackageManager):
	def __init__(self, repo_path: Path):
		self.repo_path = repo_path
		self.venv_path = repo_path / "venv"

		self._pip_in_venv = [self.venv_path / "bin" / "pip"]

	@override
	def prepend_commands(self, prefix: list[str]):
		self._pip_in_venv = prefix + self._pip_in_venv

	@override
	async def freeze_command(self):
		retCode, stdout, stderr = await subprocess_run_log_async(
			self._pip_in_venv + ["freeze"],
			logger=logger,
		)
		return stdout

	@override
	async def _create_venv_command(self, python_version: str):
		await subprocess_run_log_async(
			[
				"python" + python_version,
				"-m",
				"venv",
				self.venv_path,
			],
			logger=logger,
		)

	@override
	async def _install_dependencies_command(self, dependencies: list[str]):
		await subprocess_run_log_async(
			self._pip_in_venv
			+ [
				"install",
				"--progress-bar",
				"off",
			]
			+ dependencies,
			logger=logger,
		)

	@override
	async def _install_requirements_txt_command(self):
		await subprocess_run_log_async(
			self._pip_in_venv + ["install", "--progress-bar", "off"],
			logger=logger,
		)

	@override
	async def _install_self_command(self):
		await subprocess_run_log_async(
			self._pip_in_venv
			+ [
				"install",
				"--progress-bar",
				"off",
				"--no-build-isolation",
				self.repo_path,
			],
			logger=logger,
		)



import ai_parade._compat  # noqa: F401
import ai_parade._toolkit.datasets.loadImageNet as datasets
import ai_parade._toolkit.runtime_metadata as runtime_metadata
from ai_parade._toolkit.conversion_error import ConversionError
from ai_parade._toolkit.convert import Convert
from ai_parade._toolkit.inference import inference

from ._toolkit.capture_outputs import CaptureOutputs, LogSummary
from ._toolkit.converters.converter import ConversionResult
from ._toolkit.data.ImageInput import ImageInput
from ._toolkit.data.output import ModelOutput, Output
from ._toolkit.datasets.Dataset import Dataset
from ._toolkit.enums import (
	COMMON_HW_ACCELERATIONS,
	AIRuntime,
	ExactModelFormat,
	HardwareAcceleration,
	ModelFormat,
	Platform,
	Quantization,
	format_suffixes,
	formats_preferred_runtime,
)
from ._toolkit.get_weights import get_weights_name
from ._toolkit.logging import get_ai_parade_logger
from ._toolkit.metadata.custom import Customizations
from ._toolkit.metadata.enums import ModelTasks
from ._toolkit.metadata.models.final import (
	InstallOptions,
	ModelMetadata,
	PyTorchOptions,
)
from ._toolkit.metadata.parsing import (
	MetadataNotFoundError,
	get_local_metadata,
	parse_metadata,
)
from ._toolkit.run.ModelRunner import LoadedModelRunner
from ._toolkit.run.RemoteRunner import RemoteError
from ._toolkit.run.runner_selector import ModelRunner, get_runner
from ._toolkit.verify import compute_statistics

__all__ = [
	"runtime_metadata",
	"ModelMetadata",
	"InstallOptions",
	"PyTorchOptions",
	"ModelTasks",
	"ModelFormat",
	"AIRuntime",
	"formats_preferred_runtime",
	"ModelOutput",
	"Output",
	"ImageInput",
	"Dataset",
	"datasets",
	"CaptureOutputs",
	"LogSummary",
	"get_runner",
	"LoadedModelRunner",
	"get_local_metadata",
	"parse_metadata",
	"MetadataNotFoundError",
	"RemoteError",
	"ModelRunner",
	"Customizations",
	"get_ai_parade_logger",
	"compute_statistics",
	"Quantization",
	"ExactModelFormat",
	"ConversionResult",
	"Convert",
	"ConversionError",
	"HardwareAcceleration",
	"COMMON_HW_ACCELERATIONS",
	"format_suffixes",
	"Platform",
	"get_weights_name",
	"inference",
]



from pathlib import Path
from typing import Optional

from pydantic import ConfigDict, DirectoryPath

from ai_parade._toolkit.conversion_error import ConversionError
from ai_parade._toolkit.converters.ConversionResult import ConversionResult

from .api import InstallOptionsApi, ModelMetadataApi, PyTorchOptionsApi


class ModelMetadata(ModelMetadataApi):
	"""
	Model metadata -- everything needed to install and run the model and more
	"""

	model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

	repository_path: Optional[DirectoryPath] = None
	"""
	Path to the repository, or None if the model is not locally available
	"""

	pytorch: Optional["PyTorchOptions"] = None  # type: ignore
	install: Optional["InstallOptions"] = None  # type: ignore

	def with_conversion_results(
		self,
		conversion_result_path: Path,
		conversion_result: ConversionResult | None = None,
	) -> tuple["ModelMetadata", Path]:
		conversion_result = conversion_result or ConversionResult.load(
			conversion_result_path
		)
		if conversion_result.status == "failed":
			raise ConversionError(
				self.format,
				conversion_result.format,
				"Using failed conversion result",
			)

		new_metadata = self.model_copy(
			update={
				"image_input": conversion_result.image_input,
				"format": conversion_result.format,
				"output": conversion_result.output,
				"get_custom_runner": None,
				"get_installer": None,
			}
		)
		return new_metadata, (conversion_result_path / conversion_result.path).resolve()


class InstallOptions(InstallOptionsApi):
	"""
	Installation options
	"""

	model_config = ConfigDict(frozen=True)
	build_system_dependencies: list[str]
	"""
	Build system dependencies from pyproject (build-system.requires)
	"""
	python: str
	"""
	The environment is initialized with python of this version.
	
	"""


class PyTorchOptions(PyTorchOptionsApi):
	"""
	PyTorch options
	"""

	model_config = ConfigDict(frozen=True, protected_namespaces=())
	pass



from typing import override

from ai_parade._toolkit.enums import ModelFormat
from ai_parade._toolkit.run.ModelRunner import ModelRunner
from ai_parade._toolkit.run.ONNX import ONNXRunner

from .converter import ConversionResult, Converter, FormatConverter


class ONNX2TensorFlowConverter(FormatConverter):
	"""
	https://github.com/PINTO0309/onnx2tf
	"""

	@property
	@override
	def format_conversion(self):
		return (ModelFormat.ONNX, ModelFormat.SavedModel)

	@override
	def create(self, options: Converter.Options):
		super().create(options)
		return self.convert

	def convert(self, runner: ModelRunner):
		assert isinstance(runner, ONNXRunner)
		from onnx2tf import convert  # type: ignore

		convert(
			input_onnx_file_path=str(runner.weights),
			output_folder_path=str(self.get_output_weights_path(runner, None)),
			# output_signaturedefs: Optional[bool] = False,
			# output_h5: Optional[bool] = False,
			# output_keras_v3: Optional[bool] = False,
			# output_tfv1_pb: Optional[bool] = False,
			# output_weights: Optional[bool] = False,
			# copy_onnx_input_output_names_to_tflite: Optional[bool] = False,
			# output_integer_quantized_tflite: Optional[bool] = False,
			# quant_type: Optional[str] = 'per-channel',
			# custom_input_op_name_np_data_path: Optional[List] = None,
			# input_quant_dtype: Optional[str] = 'int8',
			# output_quant_dtype: Optional[str] = 'int8',
			# not_use_onnxsim: Optional[bool] = False,
			# not_use_opname_auto_generate: Optional[bool] = False,
			# batch_size: Union[int, NoneType] = None,
			# overwrite_input_shape: Union[List[str], NoneType] = None,
			# no_large_tensor: Optional[bool] = False,
			# output_nms_with_dynamic_tensor: Optional[bool] = False,
			# keep_ncw_or_nchw_or_ncdhw_input_names: Union[List[str], NoneType] = None,
			# keep_nwc_or_nhwc_or_ndhwc_input_names: Union[List[str], NoneType] = None,
			# keep_shape_absolutely_input_names: Optional[List[str]] = None,
			# input_names_to_interrupt_model_conversion: Union[List[str], NoneType] = None,
			# output_names_to_interrupt_model_conversion: Union[List[str], NoneType] = None,
			# disable_group_convolution: Union[bool, NoneType] = False,
			# enable_batchmatmul_unfold: Optional[bool] = False,
			# enable_rnn_unroll: Optional[bool] = False,
			# disable_suppression_flextranspose: Optional[bool] = False,
			# number_of_dimensions_after_flextranspose_compression: Optional[int] = 6,
			# disable_suppression_flexstridedslice: Optional[bool] = False,
			# disable_strict_mode: Optional[bool] = False,
			# number_of_dimensions_after_flexstridedslice_compression: Optional[int] = 5,
			# optimization_for_gpu_delegate: Optional[bool] = False,
			# replace_argmax_to_reducemax_and_indices_is_int64: Union[bool, NoneType] = False,
			# replace_argmax_to_reducemax_and_indices_is_float32: Union[bool, NoneType] = False,
			# replace_argmax_to_fused_argmax_and_indices_is_int64: Union[bool, NoneType] = False,
			# replace_argmax_to_fused_argmax_and_indices_is_float32: Union[bool, NoneType] = False,
			# fused_argmax_scale_ratio: Union[float, NoneType] = 0.5,
			# replace_to_pseudo_operators: List[str] = None,
			# mvn_epsilon: Union[float, NoneType] = 0.0000000001,
			# param_replacement_file: Optional[str] = '',
			# check_gpu_delegate_compatibility: Optional[bool] = False,
			# check_onnx_tf_outputs_elementwise_close: Optional[bool] = False,
			# check_onnx_tf_outputs_elementwise_close_full: Optional[bool] = False,
			# check_onnx_tf_outputs_sample_data_normalization: Optional[str] = 'norm',
			# check_onnx_tf_outputs_elementwise_close_rtol: Optional[float] = 0.0,
			# check_onnx_tf_outputs_elementwise_close_atol: Optional[float] = 1e-4,
			# disable_model_save: Union[bool, NoneType] = False,
			# non_verbose: Union[bool, NoneType] = False,
			# verbosity: Optional[str] = 'debug'
		)

		return ConversionResult(
			format=self.conversion[1],
			image_input=runner.model_metadata.image_input,
			path=self.get_output_weights_path(runner, None),
			output=runner.model_metadata.output,
		)
